import numpy as np


def array_from(d):
    return np.array(list(d.values()))


def array_to(keys, array):
    return {
        key: array[index]
        for index, key in enumerate(keys)}


def array_to_nonzero(keys, array):
    return {
        key: array[index]
        for index, key in enumerate(keys)
        if array[index] != 0}


def type_of(array):
    if len(array) == 0:
        return None

    head = array[0]
    if isinstance(head, (list, np.ndarray)):
        return type_of(head)
    else:
        return type(head)


def arrays_from(ds, keys):
    if not ds:
        return np.array([])

    arrays = {
        key: []
        for key in keys}

    for d in ds:
        for key, value in d.items():
            if key in arrays:
                arrays[key].append(value)

    return tuple([
        np.array(array, dtype=type_of(array))
        for array in arrays.values()])


def arrays_to(n, attrs):
    ds = []
    for index in np.arange(n):
        d = {}
        for attr in attrs.keys():
            d[attr] = attrs[attr][index]
        ds.append(d)

    return ds


def bulk_schema(elements):
    return {
        element: {
            '_default': 0,
            '_emit': True}
        for element in elements}


def mw_schema(mass_dict):
    return {
        element: {
            '_properties': {
                'mw': mw}}
        for element, mw in mass_dict.items()}


def listener_schema(elements):
    return {
        element: {
            '_default': default,
            '_updater': 'set',
            '_emit': True}
        for element, default in elements.items()}


def add_elements(elements, element_id):
    return {
        '_add': [{
            'key': element[element_id],
            'state': element}
            for element in elements]}



from typing import Optional, cast
import pytest

from vivarium.composites.toys import Qo, ToyProcess, ToyComposer
from vivarium.core.process import Process
from vivarium.core.engine import Engine
from vivarium.core.store import Store
from vivarium.core.control import run_library_cli


def get_toy_store() -> Store:
    """get a store to test the api with"""
    composer = ToyComposer({})
    store1 = composer.generate_store()
    composite = composer.generate()
    store2 = composite.generate_store()
    assert set(store1.inner.keys()) == set(store2.inner.keys())
    return store2


def test_insert_process(return_value: bool = False) -> Optional[Store]:
    """Test Store.insert by adding a new process

    Return:
        the toy store with a process3
    """
    store = get_toy_store()
    process = ToyProcess({'name': 'process3'})
    store.insert({
        'processes': {'process3': process},
        'topology': {
            'process3': {
                'port1': {'_path': ('store_C',), 'var_a': ('var_d',)},
                'port2': ('store_A',)}},
        'initial_state': {}})

    assert isinstance(
        store['process3'].value, Process), \
        'process3 not inserted successfully'
    if return_value:
        return store
    return None


def test_rewire_ports() -> None:
    """connect a process' ports to different store"""
    store = cast(Store, test_insert_process(return_value=True))

    # connect process1's port1 to the store at process3's port1
    store = cast(Store, test_insert_process(return_value=True))
    store['process1'].connect('port1', store['process3']['port1'])
    assert store['process1']['port1'] == store['process3']['port1']

    # connect process2's port2 to store store_A
    store = cast(Store, test_insert_process(return_value=True))
    store['process2'].connect('port2', store['store_A'])
    assert store['process2', 'port2', 'var_a'] == store['store_A', 'var_a']

    # turn variable 'var_a' into 'var_b'
    store = cast(Store, test_insert_process(return_value=True))
    store['process2'].connect(['port2', 'var_a'], store['store_A', 'var_b'])
    # store['process2', 'port2', 'var_a'] = store['store_A', 'var_b']
    assert store['process2', 'port2', 'var_a'] == store['store_A', 'var_b']

    sim = Engine(store = store)
    sim.update(1.0)


def test_embedded_rewire_ports() -> None:
    """rewire process ports embedded down in the hierarchy"""
    composer = ToyComposer({})

    # embed further down a path
    composite = composer.generate(path=('down1', 'down2'))
    store = composite.generate_store()

    # assert process2 is still connected to store_C
    assert store['down1', 'down2', 'process2', 'port2'] == \
           store['down1', 'down2', 'store_C']

    # rewire process2 port2 to store_A, and assert change of wiring
    store['down1', 'down2', 'process2'].connect(
        'port2', store['down1', 'down2', 'store_A'])
    assert store['down1', 'down2', 'process2', 'port2', 'var_a'] == \
           store['down1', 'down2', 'store_A', 'var_a']


def test_replace_process() -> None:
    """replace a process"""
    store = get_toy_store()
    process4 = ToyProcess()
    store['process4'] = process4
    assert store['process4'].value == process4

    # replace a process with different ports entirely
    store['process1'] = Qo({})

    # test if initial values are kept the same, and are not overwritten
    store.create(['A', 'a'], 11)
    store['process1'] = ToyProcess({})
    assert store['A', 'a'].value == 11

def test_disconnected_store_failure() -> None:
    """Test that inserting a Store into the tree results in an exception"""
    store = get_toy_store()

    with pytest.raises(Exception):
        store['ddd'] = Store({})
        store['process1', 'port1'] = store['store_D']

    with pytest.raises(Exception):
        store['process1', 'port1'] = Store({'_value': 'NEW STORE'})




# def test_connect_to_new_store() -> None:
#     """
#     topology before:
#     process3: {
#         'A': ('aaa',),
#         'B': ('bbb',),}
#
#     topology after:
#     process3: {
#         'A': ('ddd',),
#         'B': ('bbb',),}
#     """
#     store = get_toy_store()
#
#     # connect a new store to the tree
#     store['ddd'] = Store({})
#
#     # connect port A to the new store ddd
#     store['process2']['A'] = Store({})
#
#     assert store['process2'].topology == {'A': ('ddd',), 'B': ('ccc',)}
#     assert store['ddd']['a'] == 0


def test_set_value() -> None:
    """set a value through a port"""
    store = get_toy_store()
    store['process1']['port1']['var_a'] = 5
    assert store['process1']['port1']['var_a'].value == 5


def test_run_store_in_experiment() -> None:
    """put a store in an experiment and run it"""
    store = get_toy_store()

    # retrieve the processes and topology
    processes = store.get_processes()
    topology = store.get_topology()
    _ = processes  # set to _ to pass lint test
    _ = topology

    # run the experiment with a topology
    experiment = Engine(store=store)
    experiment.update(10)
    data = experiment.emitter.get_data()

    assert experiment.processes['process1'] == store['process1'].value
    assert experiment.processes['process2'] == store['process2'].value
    assert data[10.0] != data[0.0]

    print(data)


def test_run_inserted_store() -> None:
    """Make a store using the API, run it as a simulation"""
    store = Store({})
    store["p1"] = ToyProcess({'name': 'p1'})
    store["p2"] = ToyProcess({'name': 'p2'})
    sim = Engine(store = store)
    sim.update(1.0)


def test_run_rewired_store() -> None:
    """Make a store using the API, run it as a simulation"""
    store = Store({})
    store["p1"] = ToyProcess({'name': 'p1'})
    store["p2"] = ToyProcess({'name': 'p2'})
    store["p1"].connect(('port1',), store['p2', "port2"])
    sim = Engine(store = store)
    sim.update(1.0)


def test_divide_store() -> None:
    store = Store({})
    store.create(['top', 'process1'], ToyProcess({}))
    store.create(['top', 'store1', 'X'])
    store['top', 'process1'].connect('port1', 'store1')

    # divide store1 into two daughters
    store['top'].divide({
        'mother': 'store1',
        'daughters': [
            {'key': 'store2'},
            {'key': 'store3'}
        ]})

    final_state = store.get_value()
    assert 'store2' in final_state['top']
    assert 'store3' in final_state['top']
    assert 'store1' not in final_state['top']


def test_update_schema() -> None:
    store = Store({})
    store.create(['top', 'process1'], ToyProcess({}))
    store.create(['top', 'store1'], _updater='set')
    assert store['top', 'store1'].updater.__name__ == 'update_set', \
        'updater is not set correctly'


def test_port_connect() -> None:
    # create the root
    store = Store({})

    # create a new store at a path
    store.create(['top', 'store1'])
    store.create(['top', 'store2'])

    # create a process at a path
    store.create(['top', 'process1'], ToyProcess({}))
    store.create(['top', 'process2'], ToyProcess({}))

    # connect port using a relative path
    store['top', 'process1'].connect('port1', 'store1')

    # connect using store target through a different port
    store['top', 'process1'].connect('port2', store['top', 'process1', 'port1'])

    # connect using absolute path
    store['top', 'process1'].connect('port2', ('top', 'store2'), absolute=True)

    assert store['top', 'process1'].topology == {
        'port1': ('store1',), 'port2': ('store2',)}


def test_add_store() -> None:
    store = get_toy_store()

    v = store.get_value()
    assert len(v) == 5

    store.add({
        'key': '_unique_id',
        'state': {'var_a': 1, 'var_b': 2}
    })
    v = store.get_value()
    assert len(v) == 6

    store.add({
        'key': 'store_D',
        'state': {'var_a': 1, 'var_b': 2}
    })

    v = store.get_value()
    assert len(v) == 7

    # test that you cannot add a store with the same name
    with pytest.raises(Exception):
        store.add({
            'key': 'store_D',
            'state': {'var_c': 100}})


test_library = {
    '1': test_insert_process,
    '2': test_rewire_ports,
    '3': test_embedded_rewire_ports,
    '4': test_replace_process,
    '5': test_disconnected_store_failure,
    # '6': test_connect_to_new_store,
    '7': test_set_value,
    '8': test_run_store_in_experiment,
    '9': test_divide_store,
    '10': test_update_schema,
    '11': test_port_connect,
    '12': test_add_store,
    '13': test_run_inserted_store,
    '14': test_run_rewired_store,
}


# python vivarium/experiments/store_api.py -n [test number]
if __name__ == '__main__':
    run_library_cli(test_library)



import copy
import math

from scipy import constants
import numpy as np

from vivarium.core.process import Deriver
from vivarium.library.units import units
from vivarium.library.dict_utils import deep_merge



PI = math.pi
AVOGADRO = constants.N_A * 1 / units.mol



def length_from_volume(volume, width):
    '''
    get cell length from volume, using the following equation for capsule volume, with V=volume, r=radius,
    a=length of cylinder without rounded caps, l=total length:

    V = (4/3)*PI*r^3 + PI*r^2*a
    l = a + 2*r
    '''
    radius = width / 2
    cylinder_length = (volume - (4/3) * PI * radius**3) / (PI * radius**2)
    total_length = cylinder_length + 2 * radius
    return total_length

def volume_from_length(length, width):
    '''
    inverse of length_from_volume
    '''
    radius = width / 2
    cylinder_length = length - width
    volume = cylinder_length * (PI * radius**2) + (4 / 3) * PI * radius**3
    return volume

def surface_area_from_length(length, width):
    '''
    SA = 3*PI*r^2 + 2*PI*r*a
    '''
    radius = width / 2
    cylinder_length = length - width
    surface_area = 3 * PI * radius**2 + 2 * PI * radius * cylinder_length
    return surface_area



class EcoliShape(Deriver):
    """
    Process for deriving volume, mmol_to_counts, and shape from the cell mass
    """

    name = 'globals_deriver'
    defaults = {
        'width': 1,  # um
        'initial_mass': 1400 * units.fg,  # wet mass in fg
        'density': 1100 * units.g / units.L,
        'periplasm_volume_fraction': 0.3,
    }

    def __init__(self, parameters=None):
        super().__init__(parameters)

    def initial_state(self, config=None):
        volume = self.parameters['initial_mass'] / self.parameters['density']
        mmol_to_counts = (AVOGADRO * volume)
        return {
            'global': {
                'volume': volume.to('fL'),
                'mmol_to_counts': mmol_to_counts.to('L/mmol')}}

    def ports_schema(self):
        set_states = [
            'volume', 'mmol_to_counts', 'length', 'surface_area',
            'periplasm_volume',
        ]
        split_divide = [
            'volume', 'length', 'surface_area', 'periplasm_volume']
        emit = {'global': ['volume', 'width', 'length', 'surface_area']}

        # default state
        mass = self.parameters['initial_mass']
        width = self.parameters['width']
        density = self.parameters['density']
        volume = mass/density
        mmol_to_counts = (AVOGADRO * volume).to('L/mmol')
        length = length_from_volume(volume.magnitude, width)
        surface_area = surface_area_from_length(length, width)
        periplasm_volume = volume * self.parameters[
            'periplasm_volume_fraction']

        default_state = {
            'global': {
                'mass': mass,
                'volume': volume.to('fL'),
                'mmol_to_counts': mmol_to_counts,
                'density': density,
                'width': width,
                'length': length,
                'surface_area': surface_area,
                'periplasm_volume': periplasm_volume,
            },
        }

        schema = {}
        for port, states in default_state.items():
            schema[port] = {}
            for state_id, value in states.items():
                schema[port][state_id] = {}
                if state_id in set_states:
                    schema[port][state_id]['_updater'] = 'set'
                if state_id in emit[port]:
                    schema[port][state_id]['_emit'] = True
                if state_id in split_divide:
                    schema[port][state_id]['_divider'] = 'split'
                if state_id in default_state[port]:
                    schema[port][state_id]['_default'] = default_state[port][state_id]

        return schema

    def next_update(self, timestep, states):
        density = states['global']['density']
        mass = states['global']['mass'].to('fg')
        width = states['global']['width']

        # get volume from mass, and more variables from volume
        volume = mass / density
        mmol_to_counts = (AVOGADRO * volume).to('L/mmol')
        length = length_from_volume(volume.magnitude, width)
        surface_area = surface_area_from_length(length, width)
        periplasm_volume = volume * self.parameters[
            'periplasm_volume_fraction']

        return {
            'global': {
                'volume': volume.to('fL'),
                'mmol_to_counts': mmol_to_counts,
                'length': length,
                'surface_area': surface_area,
                'periplasm_volume': periplasm_volume,
            },
        }



def test_deriver(total_time=10, return_value=False):

    growth_rate = 6e-3

    # configure process
    deriver = EcoliShape({})
    state = deriver.default_state()

    # initialize saved data
    saved_state = {}

    ## simulation
    timeline = [(total_time, {})]
    time = 0
    timestep = 1  # sec
    saved_state[time] = state
    while time < timeline[-1][0]:
        time += timestep
        for (t, change_dict) in timeline:
            if time >= t:
                for key, change in change_dict.items():
                    state[key].update(change)

        # set mass, counts
        mass = state['global']['mass']
        state['global']['mass'] = mass * np.exp(growth_rate * timestep)

        # get update
        update = deriver.next_update(timestep, state)

        # set derived state
        state['global'].update(update['global'])

        # save state
        saved_state[time] = copy.deepcopy(state)

    if return_value:
        return saved_state


if __name__ == '__main__':
    saved_data = test_deriver(100, return_value=True)
    print(saved_data)



"""
===============
Alternator Process
===============
"""

import os

from typing import Any, Dict

from vivarium.core.engine import pp, Engine
from vivarium.core.process import (
    Process,
    Deriver
)
from vivarium.core.composer import Composer
from vivarium.core.directories import PROCESS_OUT_DIR
from vivarium.composites.toys import ExchangeA


NAME = 'alternator'


def find_chosen(choices):
    '''
    given a dict of choices key -> True/False where only one value is True,
    return the index of the True value.
    '''

    for index, option in enumerate(choices.items()):
        choice, chosen = option
        if chosen:
            return choice, index
    return None, -1


def choose_option(choices, chosen_index):
    '''
    given a list of choices and an index, construct a dict where
    the keys are the choices and the values are False everywhere
    except the given index
    '''

    return {
        choice: index == chosen_index
        for index, choice in enumerate(choices)}


class PeriodicEvent(Process):
    """ PeriodicEvent Process

    Alternate between different choices for the given periods of time.
    """

    defaults: Dict[str, Any] = {
        'periods': [],
    }

    def __init__(self, parameters=None):
        super().__init__(parameters)
        self.periods = self.parameters['periods']

    def calculate_timestep(self, states):
        num_periods = len(self.parameters['periods'])
        period_index = states['period_index'] % num_periods
        period = self.parameters['periods'][period_index]

        return period

    def ports_schema(self):
        return {
            'event_trigger': {
                '_default': False,
                '_updater': 'set'},
            'period_index': {
                '_default': 0,
                '_updater': 'set'}}

    def next_update(self, timestep, states):
        print('in periodic_event next_update')

        num_periods = len(self.parameters['periods'])
        next_index = (states['period_index'] + 1) % num_periods
        return {
            'event_trigger': True,
            'period_index': next_index}


class Alternator(Deriver):
    """ Alternator Process

    Alternate between different choices for the given periods of time.
    """

    defaults: Dict[str, Any] = {
        'choices': [],
    }

    def __init__(self, parameters=None):
        super().__init__(parameters)
        self.choices = self.parameters['choices']

    def calculate_timestep(self, states):
        period = 0
        num_periods = len(self.parameters['periods'])
        for index, choice in enumerate(self.parameters['choices']):
            if states[choice]:
                period = self.parameters['periods'][index % num_periods]
                break
        return period

    def initial_state(self, config=None):
        return {
            choice: index == 0
            for index, choice in enumerate(self.parameters['choices'])}

    def ports_schema(self):
        choices = {
            choice: {
                '_default': False}
            for choice in self.parameters['choices']}

        return {
            'alternate_trigger': {
                '_default': False,
                '_updater': 'set'},
            'choices': choices}

    def next_update(self, timestep, states):
        print('in alternator next_update')

        if states['alternate_trigger']:
            _, index = find_chosen(states['choices'])
            if index == -1:
                index = 0
            next_choice = (index + 1) % len(self.parameters['choices'])
            choices_update = choose_option(
                self.parameters['choices'],
                next_choice)

            return {
                'alternate_trigger': False,
                'choices': choices_update}
        return {}


# test
class ExchangeAlternator(Composer):
    defaults = {
        'on_exchange': {
            'uptake_rate': 0.2,
            '_condition': ('ON',)},
        'off_exchange': {
            'uptake_rate': 0.5,
            '_condition': ('OFF',)},
        'periodic': {
            'periods': [10.0, 4.0]},
        'alternator': {
            'choices': ['ON', 'OFF']}}

    def generate_processes(self, config):
        return {
            'on_exchange': ExchangeA(config['on_exchange']),
            'off_exchange': ExchangeA(config['off_exchange']),
            'periodic': PeriodicEvent(config['periodic']),
            'alternator': Alternator(config['alternator'])}

    def generate_topology(self, config):
        return {
            'on_exchange': {
                'internal': ('concentrations', 'ON'),
                'external': ('concentrations', 'OFF')},
            'off_exchange': {
                'internal': ('concentrations', 'OFF'),
                'external': ('concentrations', 'ON')},
            'periodic': {
                'event_trigger': ('trigger',),
                'period_index': ('period_index',)},
            'alternator': {
                'alternate_trigger': ('trigger',),
                'choices': {
                    'ON': ('ON',),
                    'OFF': ('OFF',)}}}


def test_alternator(return_value=False):
    initial_on = 10.0
    initial_off = 13.0

    initial_state = {
        'concentrations': {
            'ON': {
                'A': initial_on},
            'OFF': {
                'A': initial_off}
        },
        'ON': False,
        'OFF': True,
    }

    # declare the hierarchy
    experiment = Engine(
        composite=ExchangeAlternator().generate(),
        initial_state=initial_state
    )

    pp(experiment.topology)
    pp(experiment.state.get_value())

    # timeline triggers alternator for agent_s
    time_total = 25.0

    # run simulation
    experiment.update(time_total)
    output = experiment.emitter.get_data()
    experiment.end()  # end required for parallel processes

    if return_value:
        return output
    return None


def run_alternator():
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    os.makedirs(out_dir, exist_ok=True)
    output = test_alternator(return_value=True)
    pp(output)


if __name__ == '__main__':
    run_alternator()



import csv
import io
import os
from typing import Dict, List, Optional, Any

import numpy as np

from vivarium.library.dict_utils import flatten_timeseries


def process_path_timeseries_for_csv(path_ts):
    # type: (dict) -> dict
    '''Prepare path timeseries data for writing to CSV

    Processing Steps:

    1. Convert any dictionary keys that are tuples to strings by joining
       the tuple elements with ``-``.
    2. Remove from the timeseries any data where each timepoint is not
       numeric. For example, we remove data where each timepoint is a
       matrix or a string.

    .. note:: We assume that across timepoints for the same variable,
        the data types are the same.

    Returns:
        dict: A timeseries that can be saved to a CSV with
        :py:func:`save_flat_timeseries`.
    '''
    # Convert tuple keys to strings
    str_keys = dict()
    for key, value in path_ts.items():
        try:
            iter(key)
            if not isinstance(key, str):
                key = ",".join(key)
        except TypeError:
            pass
        str_keys[key] = value

    remove_keys = [
        # Remove matrices
        key for key, val in str_keys.items()
        if isinstance(val[0], list)
    ]
    # Remove keys with non-numeric data
    for key, val in str_keys.items():
        try:
            float(val[0])
        except (ValueError, TypeError):
            remove_keys.append(key)
    for key in set(remove_keys):
        del str_keys[key]
    return str_keys


def save_flat_timeseries(timeseries, out_dir, filename):
    # type: (dict, str, str) -> None
    '''Save a flattened timeseries to a CSV file

    The CSV file will have one column for each key in the timeseries.
    The heading will be the key, and the rows will contain the
    timeseries data, one row per timepoint, in increasing order of time.
    '''
    n_rows = max([len(val) for val in timeseries.values()])
    rows: List[Dict] = [{} for _ in range(n_rows)]
    for key, val in timeseries.items():
        for i, elem in enumerate(val):
            rows[i][key] = elem
    with open(os.path.join(out_dir, filename), 'w') as f:
        writer = csv.DictWriter(f, timeseries.keys(), delimiter=',')
        writer.writeheader()
        writer.writerows(rows)


def assert_timeseries_close(
    timeseries1, timeseries2, keys=None,
    default_tolerance=(1 - 1e-10), tolerances: Optional[Dict[str, Any]] = None,
    required_frac_checked=0.9,
):
    '''Check that two timeseries are similar.

    Ensures that each pair of data points between the two timeseries are
    within a tolerance of each other, after filtering out timepoints not
    common to both timeseries.

    Arguments:
        timeseries1: One timeseries. Must be flattened and include times
            under the 'time' key.
        timeseries2: The other timeseries. Same requirements as
            timeseries1.
        keys: Keys of the timeseries whose values will be checked for
            correlation. If not specified, all keys present in both
            timeseries are used.
        default_tolerance: The tolerance to use when not specified in
            tolerances.
        tolerances: Dictionary of key-value pairs where the key is a key
            in both timeseries and the value is the tolerance to use
            when checking that key.
        required_frac_checked: The required fraction of timepoints in a
            timeseries that must be checked. If this requirement is not
            satisfied, which might occur if the two timeseries share few
            timepoints, the test wll fail.

    Raises:
        AssertionError: If a pair of data points have a difference
            strictly above the tolerance threshold or if too few
            timepoints are common to both timeseries.
    '''
    tolerances = tolerances or {}
    arrays1, arrays2, keys = _prepare_timeseries_for_comparison(
        timeseries1, timeseries2, keys, required_frac_checked)
    for key in keys:
        tolerance = tolerances.get(key, default_tolerance)
        close_mask = np.isclose(
            arrays1[key], arrays2[key], atol=tolerance, equal_nan=True)
        if not np.all(close_mask):
            print('Timeseries 1:', arrays1[key][~close_mask])
            print('Timeseries 2:', arrays2[key][~close_mask])
            raise AssertionError(
                'The data for {} differed by more than {}'.format(
                    key, tolerance)
            )


def assert_timeseries_correlated(
    timeseries1, timeseries2, keys=None,
    default_threshold=(1 - 1e-10), thresholds: Optional[Dict[str, Any]] = None,
    required_frac_checked=0.9,
):
    '''Check that two timeseries are correlated.

    Uses a Pearson correlation coefficient. Only the data from
    timepoints common to both timeseries are compared.

    Arguments:
        timeseries1: One timeseries. Must be flattened and include times
            under the 'time' key.
        timeseries2: The other timeseries. Same requirements as
            timeseries1.
        keys: Keys of the timeseries whose values will be checked for
            correlation. If not specified, all keys present in both
            timeseries are used.
        default_threshold: The threshold correlation coefficient to use
            when a threshold is not specified in thresholds.
        thresholds: Dictionary of key-value pairs where the key is a key
            in both timeseries and the value is the threshold
            correlation coefficient to use when checking that key
        required_frac_checked: The required fraction of timepoints in a
            timeseries that must be checked. If this requirement is not
            satisfied, which might occur if the two timeseries share few
            timepoints, the test wll fail. This is also the fraction of
            timepoints for each variable that must be non-nan in both
            timeseries. Note that the denominator of this fraction is
            the number of shared timepoints that are non-nan in either
            of the timeseries.

    Raises:
        AssertionError: If a correlation is strictly below the
            threshold or if too few timepoints are common to both
            timeseries.
    '''
    thresholds = thresholds or {}
    arrays1, arrays2, keys = _prepare_timeseries_for_comparison(
        timeseries1, timeseries2, keys, required_frac_checked)
    for key in keys:
        both_nan = np.isnan(arrays1[key]) & np.isnan(arrays2[key])
        valid_indices = ~(
            np.isnan(arrays1[key]) | np.isnan(arrays2[key]))
        frac_checked = valid_indices.sum() / (~both_nan).sum()
        if frac_checked < required_frac_checked:
            raise AssertionError(
                'Timeseries share too few non-nan values for variable '
                '{}: {} < {}'.format(
                    key, frac_checked, required_frac_checked
                )
            )
        corrcoef = np.corrcoef(
            arrays1[key][valid_indices],
            arrays2[key][valid_indices],
        )[0][1]
        threshold = thresholds.get(key, default_threshold)
        if corrcoef < threshold:
            raise AssertionError(
                'The correlation coefficient for '
                '{} is too small: {} < {}'.format(
                    key, corrcoef, threshold)
            )


def _prepare_timeseries_for_comparison(
    timeseries1, timeseries2, keys=None,
    required_frac_checked=0.9,
):
    '''Prepare two timeseries for comparison

    Arguments:
        timeseries1: One timeseries. Must be flattened and include times
            under the 'time' key.
        timeseries2: The other timeseries. Same requirements as
            timeseries1.
        keys: Keys of the timeseries whose values will be checked for
            correlation. If not specified, all keys present in both
            timeseries are used.
        required_frac_checked: The required fraction of timepoints in a
            timeseries that must be checked. If this requirement is not
            satisfied, which might occur if the two timeseries share few
            timepoints, the test wll fail.

    Returns:
        A tuple of an ndarray for each of the two timeseries and a list of
        the keys for the rows of the arrays. Each ndarray has a row for
        each key, in the order of keys. The ndarrays have only the
        columns corresponding to the timepoints common to both
        timeseries.

    Raises:
        AssertionError: If a correlation is strictly below the
            threshold or if too few timepoints are common to both
            timeseries.
    '''
    if 'time' not in timeseries1 or 'time' not in timeseries2:
        raise AssertionError('Both timeseries must have key "time"')
    if keys is None:
        keys = set(timeseries1.keys()) & set(timeseries2.keys())
    else:
        if 'time' not in keys:
            keys.append('time')
    keys = list(keys)
    shared_times = set(timeseries1['time']) & set(timeseries2['time'])
    frac_timepoints_checked = (
        len(shared_times)
        / min(len(timeseries1['time']), len(timeseries2['time']))
    )
    if frac_timepoints_checked < required_frac_checked:
        raise AssertionError(
            'The timeseries share too few timepoints: '
            '{} < {}'.format(
                frac_timepoints_checked, required_frac_checked)
        )
    masked = []
    for ts in (timeseries1, timeseries2):
        arrays_dict = timeseries_to_ndarrays(ts, keys)
        arrays_dict_shared_times = {}
        for key, array in arrays_dict.items():
            # Filters out times after data ends
            times_for_array = arrays_dict['time'][:len(array)]
            arrays_dict_shared_times[key] = array[
                np.isin(times_for_array, list(shared_times))]
        masked.append(arrays_dict_shared_times)
    return (
        masked[0],
        masked[1],
        keys,
    )


def timeseries_to_ndarrays(timeseries, keys=None):
    '''After filtering by keys, convert timeseries to dict of ndarrays

    Returns:
        dict: Mapping from timeseries variables to an ndarray of the
            variable values.
    '''
    if keys is None:
        keys = timeseries.keys()
    return {
        key: np.array(timeseries[key], dtype=float) for key in keys}


def load_timeseries(path_to_csv):
    '''Load a timeseries saved as a CSV using save_timeseries.

    The timeseries is returned in flattened form.
    '''
    with io.open(path_to_csv, 'r', newline='') as f:
        reader = csv.DictReader(f)
        timeseries = {}
        for row in reader:
            for header, elem in row.items():
                if elem == '':
                    elem = None
                if elem is not None:
                    elem = float(elem)
                timeseries.setdefault(header, []).append(elem)
    return timeseries


def save_timeseries(timeseries, out_dir='out', filename='timeseries.csv'):
    flattened = flatten_timeseries(timeseries)
    save_flat_timeseries(flattened, out_dir, filename)


def agent_timeseries_from_data(data, agents_key='cells'):
    timeseries = {}
    for time, all_states in data.items():
        agent_data = all_states[agents_key]
        for agent_id, ports in agent_data.items():
            if agent_id not in timeseries:
                timeseries[agent_id] = {}
            for port_id, states in ports.items():
                if port_id not in timeseries[agent_id]:
                    timeseries[agent_id][port_id] = {}
                for state_id, state in states.items():
                    if state_id not in timeseries[agent_id][port_id]:
                        timeseries[agent_id][port_id][state_id] = []
                    timeseries[agent_id][port_id][state_id].append(state)
    return timeseries



"""Plot topologies using networkx and matplotlib."""

import os
import copy
import argparse
from typing import Any, cast, Dict, Optional

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.path import Path
from matplotlib.figure import Figure
import networkx as nx

from vivarium.core.process import Process
from vivarium.core.composer import Composer, Composite
from vivarium.core.store import generate_state
from vivarium.library.topology import normalize_path
from vivarium.library.dict_utils import deep_merge
from vivarium.processes.engulf import ToyAgent
from vivarium.processes.timeline import TimelineProcess


def construct_storage_path() -> Path:
    """Construct a Path to draw the standard "storage" flowchart shape."""
    # NOTE: After a MOVETO, we need to put the pen down for CLOSEPOLY to
    # complete a filled shape.
    _path_data = [
        # main shape
        (Path.MOVETO, [-1.000, -0.800]),  # start at bottom left corner
        (Path.CURVE4, [-0.900, -1.000]),  # bottom curve
        (Path.CURVE4, [+0.900, -1.000]),
        (Path.CURVE4, [+1.000, -0.800]),
        (Path.LINETO, [+1.000, +0.800]),  # right side
        (Path.CURVE4, [+0.900, +1.000]),  # top back curve
        (Path.CURVE4, [-0.900, +1.000]),
        (Path.CURVE4, [-1.000, +0.800]),
        (Path.LINETO, [-1.000, -0.800]),  # back to the bottom
        (Path.CLOSEPOLY, [0.00, 0.00]),  # close a filled poly-line shape

        # front edge 1
        (Path.MOVETO, [-1.000, +0.800]),  # to the top left corner
        (Path.CURVE4, [-0.900, +0.600]),  # top front main curve
        (Path.CURVE4, [+0.900, +0.600]),
        (Path.CURVE4, [+1.000, +0.800]),

        # front edge 2
        (Path.MOVETO, [-1.000, +0.700]),  # to the top left corner
        (Path.CURVE4, [-0.900, +0.500]),
        (Path.CURVE4, [+0.900, +0.500]),  # top front second curve
        (Path.CURVE4, [+1.000, +0.700]),
    ]
    _path_codes, _path_vertices = zip(*_path_data)
    return Path(_path_vertices, _path_codes)


STORAGE_PATH = construct_storage_path()


def get_bigraph(composite):
    """ Get a graph with Processes, Stores, and edges from a Vivarium topology """
    topology = composite['topology']
    processes = composite['processes']
    steps = composite['steps']
    hierarchy_object = generate_state(
        processes=processes,
        topology=topology,
        initial_state={},
        steps=steps)

    # get path to processes and stores, name them by their paths
    # leaf_paths = hierarchy_object.depth(filter_function=lambda x: x.inner == {})
    process_paths = hierarchy_object.depth(filter_function=lambda x: isinstance(x.value, Process))
    process_nodes = ['\n'.join(process_path[0]) for process_path in process_paths]
    store_paths = hierarchy_object.depth(filter_function=lambda x: x.inner != {})
    store_nodes = ['\n'.join(store_path[0]) for store_path in store_paths if store_path[0] != tuple()]

    # get the edges between processes and stores
    edges = {}
    for (process_path, process) in process_paths:
        process_id = '\n'.join(process_path)
        assert process_id in process_nodes, (
            f"{process_id} process id is not in process_nodes list: {process_nodes}")

        process_topology = process.topology
        for port, store_path in process_topology.items():
            if isinstance(store_path, dict):
                if '_path' in store_path:
                    store_path = store_path['_path']
                else:
                    store_paths = tuple()

            store_path = normalize_path(process_path[:-1] + store_path)
            store_id = '\n'.join(store_path)
            if store_id not in store_nodes:
                # print(f"Adding {store_id} to store_nodes list {store_nodes}")
                store_paths.append((store_path, hierarchy_object.get_path(path=store_path)))
                store_nodes.append(store_id)

            # save the edge
            edge = (process_id, store_id)
            edges[edge] = port

    # get the place edges between hierarchy stores
    place_edges = []
    for (store_path, _) in store_paths:
        if len(store_path) > 1:
            # hierarchy place edges between inner/outer stores
            # for store_1, store_2 in zip(store_path, store_path[1:]):
            for level, _ in enumerate(store_path[1:], 1):
                store_1 = '\n'.join(store_path[:level])
                store_2 = '\n'.join(store_path[:level + 1])
                place_edge = (store_1, store_2)
                place_edges.append(place_edge)

    # are there overlapping names?
    overlap = [name for name in process_nodes if name in store_nodes]
    if overlap:
        print('{} shared by processes and stores'.format(overlap))

    return process_nodes, store_nodes, edges, place_edges


# def replace_node_labels(
#         node_labels, node_list=None, node_dict=None, edge_list=None, edge_dict=None):
#     if node_list:
#         return [
#             node_labels.get(node_id, node_id)
#             for node_id in node_list]
#     elif node_dict:
#         return {
#             node_labels.get(node_id, node_id): value
#             for node_id, value in node_dict.items()}
#     elif edge_list:
#         return [(
#             node_labels.get(node_1, node_1),
#             node_labels.get(node_2, node_2)
#         ) for (node_1, node_2) in edge_list]
#     elif edge_dict:
#         return {(
#              node_labels.get(node_1, node_1),
#              node_labels.get(node_2, node_2)
#          ): edge_name for (node_1, node_2), edge_name in edge_dict.items()}
#     return None


def remove(
        remove_nodes, node_list=None, node_dict=None, edge_list=None, edge_dict=None):
    """remove specified nodes"""
    if node_list:
        return [
            node_id for node_id in node_list
            if node_id not in remove_nodes]
    elif node_dict:
        return {
            node_id: value
            for node_id, value in node_dict.items()
            if node_id not in remove_nodes}
    elif edge_list:
        return [
            (node_1, node_2)
            for (node_1, node_2) in edge_list
            if (node_1 not in remove_nodes and node_2 not in remove_nodes)]
    elif edge_dict:
        return {
            (node_1, node_2): port
            for (node_1, node_2), port in edge_dict.items()
            if (node_1 not in remove_nodes and node_2 not in remove_nodes)}
    return None


def get_networkx_graph(composite, remove_nodes=None):
    """ Make a networkX graph from a Vivarium topology """
    remove_nodes = remove_nodes or []

    # get the nodes and edges from the composite
    process_nodes, store_nodes, edges, place_edges = get_bigraph(composite)

    # remove specified nodes
    process_nodes = remove(remove_nodes, node_list=process_nodes)
    store_nodes = remove(remove_nodes, node_list=store_nodes)
    edges = remove(remove_nodes, edge_dict=edges)
    place_edges = remove(remove_nodes, edge_list=place_edges)

    # make networkX graph
    g = nx.Graph()
    for node_id in process_nodes:
        g.add_node(node_id, type='Process')
    for node_id in store_nodes:
        g.add_node(node_id, type='Store')

    # add topology edges
    for (process_id, store_id), port in edges.items():
        g.add_edge(process_id, store_id, port=port)

    return g, place_edges


def graph_figure(
        graph: nx.Graph,
        *,
        coordinates: Optional[Dict] = None,
        node_labels: Optional[Dict] = None,
        graph_format: str = 'horizontal',
        place_edges: Optional[list] = None,
        show_ports: bool = True,
        store_color: Any = 'tab:blue',
        process_color: Any = 'tab:orange',
        store_colors: Optional[Dict] = None,
        process_colors: Optional[Dict] = None,
        color_edges: bool = False,
        dashed_edges: bool = False,
        edge_width: float = 2.0,
        fill_color: Any = 'w',
        node_size: float = 8000,
        font_size: int = 14,
        node_distance: float = 3.0,
        buffer: float = 0.5,
        border_width: float = 3,
        custom_widths: Optional[Dict] = None,
        label_pos: float = 0.65,
) -> Figure:
    """ Make a figure from a networkx graph.

    :param graph: the networkx. Graph to plot
    :param coordinates: (dict) a dictionary of locations for all nodes in the graph, with {'node_id': (x, y)}
    :param node_labels: (dict) a dictionary of labels for all nodes in the graph, with {'node_id': 'node_label'}
    :param graph_format: 'horizontal', 'vertical', or 'hierarchy'
    :param store_color: default color for the Store nodes; any matplotlib color value
    :param process_color: default color for the Process nodes; any matplotlib color value
    :param store_colors: (dict) specific colors for the Store nodes, mapping from store name to matplotlib color
    :param process_colors: (dict) specific colors for the Process nodes, mapping from store name to matplotlib color
    :param color_edges: color each edge between Store and Process a different color
    :param dashed_edges: edges between Store and Process are dashed lines
    :param show_ports: whether to show the Port labels
    :param fill_color: fill color for the Store and Process nodes; any
        matplotlib color value
    :param node_size: size to draw the Store and Process nodes
    :param font_size: size for the Store, Process, and Port labels
    :param node_distance: distance to spread out the nodes
    :param buffer: buffer space around the graph
    :param border_width: width of the border line around Store and Process nodes
    :param custom_widths: (dict) changes the widths of specific Store and Process nodes (defaults to board_width)
    :param label_pos: position of the Port labels along their connection lines,
        (0=head, 0.5=center, 1=tail)
    """
    node_labels = node_labels or {}
    process_colors = process_colors or {}
    store_colors = store_colors or {}
    place_edges = place_edges or []
    custom_widths = custom_widths or {}

    node_attributes = dict(graph.nodes.data())
    process_nodes = [
        node_id for node_id, attributes in node_attributes.items()
        if attributes['type'] == 'Process']
    store_nodes = [
        node_id for node_id, attributes in node_attributes.items()
        if attributes['type'] == 'Store']

    # fill in all node labels
    node_labels = {
        node_id: node_labels.get(node_id, node_id)
        for node_id in node_attributes.keys()}

    edge_list = list(graph.edges)
    edges = {}
    for edge in edge_list:
        if 'port' in graph.edges[edge]:
            edges[edge] = graph.edges[edge]['port']

    # get position
    pos: Dict = {}
    if graph_format:
        pos_format = graph_format_location(
            graph,
            process_nodes,
            store_nodes,
            place_edges,
            graph_format)
        pos = deep_merge(pos, pos_format)
    if coordinates:
        pos_format = {
            node: np.array(coord)
            for node, coord in coordinates.items()
            if node in process_nodes + store_nodes}
        pos = deep_merge(pos, pos_format)

    # initialize figure based on positions, nodes, and buffer
    pos_values = list(pos.values())
    xs = [p[0] for p in pos_values]
    ys = [p[1] for p in pos_values]
    xr = max(xs) - min(xs)
    yr = max(ys) - min(ys)
    fig, ax = plt.subplots(figsize=(
        xr * node_distance + 2 * buffer,
        yr * node_distance + 2 * buffer))

    # get node colors
    process_color_list = [
        process_colors.get(process_name, process_color)
        for process_name in process_nodes]
    store_color_list = [
        store_colors.get(store_name, store_color)
        for store_name in store_nodes]

    #get node widths
    process_width_list = [
        custom_widths.get(process_name, border_width)
        for process_name in process_nodes]

    store_width_list = [
        custom_widths.get(store_name, border_width)
        for store_name in store_nodes]

    # draw the process nodes
    nx.draw_networkx_nodes(graph, pos,
                           ax=ax,
                           nodelist=process_nodes,
                           node_color=fill_color,
                           edgecolors=process_color_list,
                           node_size=node_size,
                           linewidths=process_width_list,
                           node_shape='s')
    # draw the store nodes
    nx.draw_networkx_nodes(graph, pos,
                           ax=ax,
                           nodelist=store_nodes,
                           node_color=fill_color,
                           edgecolors=store_color_list,
                           node_size=node_size,
                           linewidths=store_width_list,
                           node_shape=cast(str, [STORAGE_PATH]))

    # edges
    edge_args: dict = {}

    # edge colors
    if color_edges:
        edge_args['edge_cmap'] = plt.get_cmap('nipy_spectral')
        edge_args['edge_color'] = list(range(1, len(edges) + 1))
        if graph_format == 'hierarchy':
            edge_args['edge_color'].extend([0 for _ in place_edges])

    # edge width
    edge_args['width'] = [edge_width for _ in edges.keys()]
    # thicker edges for hierarchy connections
    edge_args['width'].extend([edge_width * 2 for _ in place_edges])
    if dashed_edges:
        edge_args['style'] = ['dashed' for _ in edges.keys()]
        edge_args['style'].extend(['solid' for _ in place_edges])

    nx.draw_networkx_edges(graph, pos, ax=ax, **edge_args)

    # node labels
    nx.draw_networkx_labels(graph, pos,
                            ax=ax,
                            labels=node_labels,
                            font_size=font_size)
    if show_ports:
        # edge labels
        nx.draw_networkx_edge_labels(graph, pos,
                                     ax=ax,
                                     edge_labels=edges,
                                     font_size=font_size,
                                     label_pos=label_pos)

    # add white buffer around final figure
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    ax.set_xlim(xmin - buffer, xmax + buffer)
    ax.set_ylim(ymin - buffer, ymax + buffer)
    ax.axis('off')

    return fig


def get_hierarchy_levels(graph, store_nodes, place_edges):
    # add new place edges by iterating over all place_edges
    outers = set()
    inners = set()
    for (store_1, store_2) in place_edges:
        outers.add(store_1)
        inners.add(store_2)
        graph.add_edge(store_1, store_2, place_edge=True)

    # add non-embedded nodes to outers
    all_stores = outers.union(inners)
    non_embedded = set(store_nodes).difference(all_stores)
    outers.update(non_embedded)

    # add intermediate nodes to store_nodes
    intermediate_nodes = all_stores.difference(store_nodes)
    store_nodes.extend(list(intermediate_nodes))

    # determine the hierarchy levels
    levels = []
    accounted = set()
    unaccounted = outers.union(inners)
    top_level = outers - inners
    levels.append(list(top_level))
    accounted.update(top_level)
    unaccounted = unaccounted.difference(accounted)
    while len(unaccounted) > 0:
        next_level = set()
        for (store_1, store_2) in place_edges:
            if store_1 in accounted and store_2 in unaccounted:
                next_level.add(store_2)
        levels.append(list(next_level))
        accounted.update(next_level)
        unaccounted = unaccounted.difference(accounted)

    return levels


def graph_format_location(
    graph,
    process_nodes,
    store_nodes,
    place_edges,
    graph_format,
):
    n_stores = len(store_nodes)
    n_processes = len(process_nodes)

    # get positions
    pos = {}
    if graph_format == 'hierarchy':
        levels = get_hierarchy_levels(
            graph, store_nodes, place_edges)

        # buffer makes things centered
        n_max = max([len(level) for level in levels])

        # place the process nodes according on the left
        # buffer_processes = (n_max - n_processes) / 2
        for idx, node_id in enumerate(process_nodes, 0):
            pos[node_id] = np.array([-1, -idx])

        # place the store nodes according to levels
        for level_idx, level in enumerate(levels, 0):
            level_buffer = (n_max - len(level)) / 2
            for idx, node_id in enumerate(level, 1):
                pos[node_id] = np.array([level_buffer + 0.9*idx, -1.2*level_idx])

    elif graph_format == 'vertical':
        # processes in a column, and stores in a column
        for idx, node_id in enumerate(process_nodes, 1):
            pos[node_id] = np.array([-1, -idx])
        for idx, node_id in enumerate(store_nodes, 1):
            pos[node_id] = np.array([1, -idx])

    elif graph_format == 'horizontal':
        # processes in a row, and stores in a row
        # buffer makes things centered
        n_max = max(n_stores, n_processes)
        buffer_processes = (n_max - n_processes) / 2
        buffer_stores = (n_max - n_stores) / 2

        for idx, node_id in enumerate(process_nodes, 1):
            pos[node_id] = np.array([buffer_processes + idx, 1])
        for idx, node_id in enumerate(store_nodes, 1):
            pos[node_id] = np.array([buffer_stores + idx, -1])

    return pos


def save_network(out_dir='out', filename='network'):
    os.makedirs(out_dir, exist_ok=True)
    fig_path = os.path.join(out_dir, filename)
    print(f"Writing {fig_path}")
    plt.savefig(fig_path, bbox_inches='tight')
    # plt.close()


def plot_topology(
        composite,
        settings: Optional[Dict] = None,
        out_dir: Optional[str] = None,
        filename: Optional[str] = None,
):
    """ Plot a composite's topology """

    settings = copy.deepcopy(settings) or {}
    if isinstance(composite, Composer):
        composite = composite.generate()
    elif isinstance(composite, Process):
        composite = composite.generate()
    assert 'processes' in composite and 'topology' in composite

    # make networkx graph
    remove_nodes = settings.pop('remove_nodes', [])
    g, place_edges = get_networkx_graph(composite, remove_nodes)
    settings['place_edges'] = place_edges

    # make graph figure
    fig = graph_figure(g, **settings)

    if out_dir is not None:
        # save fig
        save_network(
            out_dir=out_dir,
            filename=filename)
    return fig


# tests
class MultiPort(Process):
    name = 'multi_port'

    def ports_schema(self):
        return {
            'a': {
                'molecule': {
                    '_default': 0,
                    '_emit': True}},
            'b': {
                'molecule': {
                    '_default': 0,
                    '_emit': True}},
            'c': {
                'molecule': {
                    '_default': 0,
                    '_emit': True}}}

    def next_update(self, timestep, states):
        return {
            'a': {'molecule': 1},
            'b': {'molecule': 1},
            'c': {'molecule': 1}}


class MergePort(Composer):
    """combines both of MultiPort's ports into one store"""
    name = 'multi_port_composer'
    defaults = {
        'topology': {
            'multiport1': {
                'a': ('A',),
                'b': ('B',),
                'c': ('C',),
            },
            'multiport2': {
                'a': ('A',),
                'b': ('B',),
                'c': ('C',),
            },
        }
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.topology = self.config['topology']

    def generate_processes(self, config):
        return {
            'multiport1': MultiPort({}),
            'multiport2': MultiPort({}),
        }

    def generate_topology(self, config):
        return self.topology


def test_merge_port_topology(
        fig_name=None,
        topology=None,
        settings=None,
):

    if topology is None:
        topology = {
            'multiport1': {},
            'multiport2': {}}
    composer = MergePort({'topology': topology})

    config = {'settings': settings}
    if fig_name:
        config.update({
            'out_dir': 'out/topology',
            'filename': fig_name})

    plot_topology(composer, **config)


def merge_port_configs(topology_id):
    topology = {}
    settings = {}
    if topology_id == '0':
        pass
    elif topology_id == '1':
        # vertical graph format
        topology = {
                'multiport1': {
                    'a': ('D',),
                    'b': ('D',),
                    'c': ('D',),
                },
                'multiport2': {}}
        settings = {
            'graph_format': 'vertical',
            'process_colors': {'multiport1': 'r'},
            'store_colors': {'C': 'k'},
        }
    elif topology_id == '2':
        # default horizontal graph format
        topology = {
                'multiport1': {
                    'a': ('A', 'AA',),
                    'b': ('A', 'BB',),
                    'c': ('A', 'CC',),
                },
                'multiport2': {}}
        settings = {'dashed_edges': True}
    elif topology_id == '3':
        # hierarchy graph format
        topology = {
                'multiport1': {
                    'a': ('A', 'AA', 'AAA',),
                    'b': ('A', 'BB',),
                    'c': ('A', 'CC',),
                },
                'multiport2': {}}
        settings = {
            'graph_format': 'hierarchy',
            'store_color': 'navy'
        }
    elif topology_id == '4':
        # manual locations
        topology = {
                'multiport1': {
                    'a': ('A', 'AA',),
                    'b': ('A', 'BB', 'BBB'),
                    'c': ('A', 'CC',),
                },
                'multiport2': {}}
        settings = {
            'coordinates': {
                'multiport1': (-1, 0), 'multiport2': (-1, -2),
                'A': (1, 0), 'B': (2, 0), 'C': (3, 0),
                'A\nAA': (1, -2), 'A\nBB': (2, -2), 'A\nCC': (3, -2),
                'A\nBB\nBBB': (2, -4),
            },
            'store_color': 'navy',
            'dashed_edges': True,
            'remove_nodes': ['A\nBB', 'multiport1']
        }
    else:
        raise ValueError(f'topology_id "{topology_id}" is invalid')
    return topology, settings


def main():

    parser = argparse.ArgumentParser(description='topology')
    parser.add_argument('-t', default=None, type=str, help='topology id for test_merge_port_topology')
    parser.add_argument('-x', action='store_true', help='hierarchy id')
    args = parser.parse_args()

    if args.t:
        topology_id = str(args.t)
        topology, settings = merge_port_configs(topology_id)
        fig_name = f"topology_{topology_id}"
        test_merge_port_topology(
            fig_name=fig_name,
            topology=topology,
            settings=settings)

    elif args.x:
        agent_ids = ['1', '2']

        config = {
            'exchange': {
                'internal_path': ('concentrations',),
                'external_path': ('..', '..', 'concentrations')},
            'engulf': {
                'inner_path': ('agents',),
                'outer_path': ('..', '..', 'agents')}}
        toy_agent_composer = ToyAgent(config)
        toy_agent_1 = toy_agent_composer.generate(path=('agents', agent_ids[0]))
        toy_agent_2 = toy_agent_composer.generate(path=('agents', agent_ids[1]))

        timeline = [(3, {('agents', agent_ids[1], 'engulf-trigger'): [agent_ids[0]]})]
        timeline_composer = TimelineProcess({'timeline': timeline})
        full_composite = timeline_composer.generate()

        full_composite.merge(composite=toy_agent_1)
        full_composite.merge(composite=toy_agent_2)

        # plot topology
        agent_1_x = 0
        agent_2_x = 6
        level_1 = -1
        level_2 = -2
        level_3 = -3
        settings = {
            'coordinates': {
                # timeline
                'timeline': (3, 0),
                # agent 1
                'agents\n1\nexchange': (agent_1_x, -1), 'agents\n1\nengulf': (agent_1_x, -1.5), 'agents\n1\nexpel': (agent_1_x, -2),
                # agent 2
                'agents\n2\nexchange': (agent_2_x, -1), 'agents\n2\nengulf': (agent_2_x, -1.5), 'agents\n2\nexpel': (agent_2_x, -2),
                # 1st level stores
                'concentrations': (2, level_1), 'agents': (3, level_1), 'global': (4, level_1/2),
                # 2nd level stores
                'agents\n1': (1.75, level_2), 'agents\n2': (4.25, level_2),
                # 3rd level stores
                # agent 1
                'agents\n1\nconcentrations': (1, level_3), 'agents\n1\nengulf-trigger': (1.5, level_3),
                'agents\n1\nexpel-trigger': (2, level_3), 'agents\n1\nagents': (2.5, level_3),
                # agent 2
                'agents\n2\nconcentrations': (3.5, level_3), 'agents\n2\nengulf-trigger': (4, level_3),
                'agents\n2\nexpel-trigger': (4.5, level_3), 'agents\n2\nagents': (5, level_3),
            },
            'graph_format': 'hierarchy',
            'store_color': 'navy',
            'dashed_edges': True,
            'node_distance': 5,
            'node_labels': {
                'agents\n1': 'agent1',
                'agents\n2': 'agent2',
                'agents\n1\nexpel-trigger': 'expel',
                'agents\n1\nengulf-trigger': 'engulf',
                'agents\n1\nconcentrations': 'conc',
                'agents\n1\nagents': 'agents',
                'agents\n2\nexpel-trigger': 'expel',
                'agents\n2\nengulf-trigger': 'engulf',
                'agents\n2\nconcentrations': 'conc',
                'agents\n2\nagents': 'agents',
            }
        }
        config = {
            'settings': settings,
            'out_dir': 'out/topology',
            'filename': 'embedded'}
        plot_topology(full_composite, **config)


if __name__ == '__main__':
    main()



'''
===========================
Simulation Output Utilities
===========================
'''

import os
from typing import Any, Dict, Optional

import random
import numpy as np
import matplotlib.pyplot as plt

from vivarium.core.emitter import path_timeseries_from_embedded_timeseries
from vivarium.library.dict_utils import get_value_from_path


def set_axes(
        ax,
        show_xaxis=False,
        sci_notation=False,
        y_offset=0.0):
    '''Set up plot axes.

    Args:
        ax: The axes to set up.
        show_xaxis: Whether to show the x axis.
        sci_notation: Either ``False`` to not use scientific notation or
            an integer :math:`x` such that scientific notation will be
            used outside the range :math:`[10^{-x}, 10^x]`.
        y_offset: Horizontal distance between axis offset text
            (typically for scientific notation) and the y-axis.
    '''
    if sci_notation:
        scilimits = 4
        if isinstance(sci_notation, int):
            scilimits = sci_notation
        ax.ticklabel_format(
            style='sci',
            axis='y',
            scilimits=(-scilimits, scilimits),
            useOffset=True)
    else:
        ax.ticklabel_format(
            style='plain',
            axis='y')
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.tick_params(right=False, top=False)

    # move offset axis text (typically scientific notation)
    t = ax.yaxis.get_offset_text()
    t.set_x(-y_offset)
    if not show_xaxis:
        ax.spines['bottom'].set_visible(False)
        ax.tick_params(bottom=False, labelbottom=False)


def _save_fig_to_dir(
        fig,
        filename,
        out_dir='out/',
):
    os.makedirs(out_dir, exist_ok=True)
    fig_path = os.path.join(out_dir, filename)
    print(f"Writing {fig_path}")
    fig.savefig(fig_path, bbox_inches='tight')


def plot_simulation_output(
        timeseries_raw,
        settings: Optional[Dict[str, Any]] = None,
        out_dir=None,
        filename='simulation',
):
    '''
    Plot simulation output, with rows organized into separate columns.

    Arguments::
        timeseries (dict): This can be obtained from simulation output with convert_to_timeseries()
        settings (dict): Accepts the following keys:

            * **column_width** (:py:class:`int`): the width (inches) of
              each column in the figure
            * **max_rows** (:py:class:`int`): ports with more states
              than this number of states get wrapped into a new column
            * **remove_zeros** (:py:class:`bool`): if True, timeseries
              with all zeros get removed
            * **remove_flat** (:py:class:`bool`): if True, timeseries
              with all the same value get removed
            * **remove_first_timestep** (:py:class:`bool`): if True,
              skips the first timestep
            * **skip_ports** (:py:class:`list`): entire ports that won't
              be plotted
            * **show_state** (:py:class:`list`): with
              ``[('port_id', 'state_id')]`` for all states that will be
              highlighted, even if they are otherwise to be removed
              TODO: Obsolete?
    '''
    int_or_float = (int, np.int32, np.int64, float, np.float32, np.float64)

    settings = settings or {}
    plot_fontsize = 8
    plt.rc('font', size=plot_fontsize)
    plt.rc('axes', titlesize=plot_fontsize)

    # get settings
    column_width = settings.get('column_width', 3)
    max_rows = settings.get('max_rows', 25)
    remove_zeros = settings.get('remove_zeros', True)
    remove_flat = settings.get('remove_flat', False)
    skip_ports = settings.get('skip_ports', [])
    remove_first_timestep = settings.get('remove_first_timestep', False)

    # make a flat 'path' timeseries, with keys being path
    top_level = list(timeseries_raw.keys())
    timeseries = path_timeseries_from_embedded_timeseries(timeseries_raw)
    time_vec = timeseries.pop('time')
    if remove_first_timestep:
        time_vec = time_vec[1:]

    # remove select states from timeseries
    removed_states = set()
    for path, series in timeseries.items():
        if path[0] in skip_ports:
            removed_states.add(path)
        elif remove_flat:
            if series.count(series[0]) == len(series):
                removed_states.add(path)
        elif remove_zeros:
            if all(v == 0 for v in series):
                removed_states.add(path)
    for path in removed_states:
        del timeseries[path]

    # get figure columns
    # get length of each top-level port
    port_lengths = {}
    for path in timeseries.keys():
        if path[0] in top_level:
            if path[0] not in port_lengths:
                port_lengths[path[0]] = 0
            port_lengths[path[0]] += 1
    n_data = [length for port, length in port_lengths.items() if length > 0]
    columns = []
    for n_states in n_data:
        new_cols = n_states / max_rows
        if new_cols > 1:
            for col in range(int(new_cols)):
                columns.append(max_rows)
            mod_states = n_states % max_rows
            if mod_states > 0:
                columns.append(mod_states)
        else:
            columns.append(n_states)

    # make figure and plot
    n_cols = len(columns)
    n_rows = max(columns)
    fig = plt.figure(figsize=(n_cols * column_width, n_rows * column_width/3))
    grid = plt.GridSpec(n_rows, n_cols)
    row_idx = 0
    col_idx = 0
    for port in port_lengths.keys():

        # get this port's timeseries
        port_timeseries = {}
        for path, ts in timeseries.items():
            if path[0] is port:
                next_path = path[1:]
                if any(isinstance(item, tuple) for item in next_path):
                    next_path = tuple([
                        item[0] if isinstance(item, tuple) else item
                        for item in next_path])
                port_timeseries[next_path] = ts

        for state_id, series in sorted(port_timeseries.items()):
            if remove_first_timestep:
                series = series[1:]
            # not enough data points -- this state likely did not exist throughout the entire simulation
            if len(series) != len(time_vec):
                continue

            ax = fig.add_subplot(grid[row_idx, col_idx])  # grid is (row, column)

            if not all(isinstance(state, int_or_float) for state in series):
                # check if series is a list of ints or floats
                ax.title.set_text(str(port) + ': ' + str(state_id) + ' (non numeric)')
            else:
                # plot line at zero if series crosses the zero line
                if any(x == 0.0 for x in series) or (any(x < 0.0 for x in series) and any(x > 0.0 for x in series)):
                    zero_line = [0 for t in time_vec]
                    ax.plot(time_vec, zero_line, 'k--')

                # plot the series
                ax.plot(time_vec, series)
                if isinstance(state_id, tuple):
                    # new line for each store
                    state_id = '\n'.join(state_id)
                ax.title.set_text(str(port) + '\n' + str(state_id))

            if row_idx == columns[col_idx]-1:
                # if last row of column
                set_axes(ax, True)
                ax.set_xlabel('time (s)')
                row_idx = 0
                col_idx += 1
            else:
                set_axes(ax)
                row_idx += 1
            ax.set_xlim((time_vec[0], time_vec[-1]))

    if out_dir:
        plt.subplots_adjust(wspace=column_width/3, hspace=column_width/3)
        _save_fig_to_dir(fig, filename, out_dir)
    return fig


def get_variable_title(path):
    '''Get figure title from a variable path.

    Args:
        path: Path to the variable.

    Returns:
        String representation of the variable suitable for a figure
        title.
    '''
    var = path[-1]
    separator = '>'
    connect_path = separator.join(path[:-1])
    if isinstance(var, tuple):  # if units are included in variable
        title = f'{connect_path}: {var[0]} ({var[1]})'
    else:
        title = f'{connect_path}: {var}'
    return title


# simple plotting function
def plot_variables(
        output,
        variables,
        column_width=8,
        row_height=1.2,
        row_padding=0.8,
        linewidth=3.0,
        sci_notation=False,
        default_color='tab:blue',
        out_dir=None,
        filename='variables'
):
    '''Create a simple figure with a timeseries for every variable.

    Args:
        output: Simulation output as a map from variable names or paths
            to timeseries data. Should contain a ``time`` key whose
            value is a list of time points.
        variables: The variables to plot. May be a list of variable
            names (if simulation output keys are just variable names) or
            a dictionary with keys ``variable`` (for the variable path),
            ``color`` (for the color to use for the plot), and
            ``display`` (the variable name to display). If ``display``
            is not provided, the result of calling
            :py:func:`get_variable_title` on the variable path is used.
        column_width: Figure width.
        row_height: Height of each row. Each variable gets one row.
        row_padding: Space between rows.
        linewidth: Width of timeseries lines.
        sci_notation: Either ``False`` for no scientific notation or an
            integer :math:`x` such that scientific notation will be used
            for values outside the range :math:`[10^{-x}, 10^x]`.
        default_color: Default timeseries color.
        out_dir: Output directory.
        filename: Output filename.

    Returns:
        The figure.
    '''
    n_rows = len(variables)
    fig = plt.figure(figsize=(column_width, n_rows * row_height))
    grid = plt.GridSpec(n_rows, 1)

    time_vec = output['time']
    for row_idx, variable_definition in enumerate(variables):
        if isinstance(variable_definition, dict):
            path = variable_definition['variable']
            var_color = variable_definition.get('color', default_color)
            variable_title = variable_definition.get('display', get_variable_title(path))
        else:
            path = variable_definition
            var_color = default_color
            variable_title = get_variable_title(path)

        # get the output timeseries
        series = get_value_from_path(output, path)

        # make a new subplot
        ax = fig.add_subplot(grid[row_idx, 0])
        ax.plot(time_vec, series, linewidth=linewidth, color=var_color)
        ax.set_title(variable_title)

        # x-axis only at bottom row
        if row_idx == n_rows - 1:
            set_axes(ax, show_xaxis=True, sci_notation=sci_notation)
            ax.set_xlabel('time (s)')
            ax.spines['bottom'].set_position(('axes', -0.2))
        else:
            set_axes(ax, sci_notation=sci_notation)

    fig.subplots_adjust(hspace=row_padding)
    if out_dir:
        _save_fig_to_dir(fig, filename, out_dir)
    return fig




if __name__ == '__main__':
    total_time = 500
    data = {
        'x': [random.uniform(0, 10) for _ in range(total_time)],
        'y': [np.exp(random.uniform(0, 10)) for _ in range(total_time)],
        'time': [t for t in range(total_time)]
    }

    plot_variables(
        output=data,
        variables=['x', 'y'],
        out_dir='out')



"""
Experiment to test maximum BSON document size with MongoDB emitter
"""
import uuid
import random

import numpy as np

from vivarium.core.engine import Engine
from vivarium.core.process import Process
from vivarium.core.composer import Composer
from vivarium.core.emitter import (
    get_experiment_database,
    data_from_database,
    data_to_database,
    delete_experiment_from_database)


class ManyParametersProcess(Process):
    defaults = {
        'number_of_parameters': 100,
        'arr_size': 1}
    def __init__(self, parameters=None):
        # make a bunch of parameters
        random_parameters = {
            key: random.random()
            for key in range(parameters['number_of_parameters'])}
        super().__init__({**parameters, **random_parameters})
    def ports_schema(self):
        return {'port': {'_default': np.random.random(
            self.parameters['arr_size']), '_emit': True}}
    def next_update(self, timestep, states):
        return {'port': 1}


class ManyParametersComposite(Composer):
    defaults = {
        'number_of_processes': 10,
        'number_of_parameters': 100,
        'arr_size': 1}
    def __init__(self, config=None):
        super().__init__(config)
        self.process_ids = [
            f'process_{key}'
            for key in range(self.config['number_of_processes'])]
    def generate_processes(self, config):
        # make a bunch of processes
        return {
            process_id: ManyParametersProcess({
                'name': process_id,
                'number_of_parameters': self.config['number_of_parameters'],
                'arr_size': self.config['arr_size']})
            for process_id in self.process_ids}
    def generate_topology(self, config):
        return {
            process_id: {'port': ('store', process_id,)}
            for process_id in self.process_ids}


def test_large_emits():
    """
    This experiment runs a large experiment to test the database emitter.
    This requires MongoDB to be configured and running.
    """
    config = {
        'number_of_processes': 10,
        'number_of_parameters': 100000,
        'arr_size': 150000}

    composer = ManyParametersComposite(config)
    composite = composer.generate()

    settings = {
        'experiment_name': 'large database experiment',
        'experiment_id': f'large_{str(uuid.uuid4())}',
        'emitter': 'database',
        'emit_processes': True
    }

    experiment = Engine(**{
        'processes': composite['processes'],
        'topology': composite['topology'],
        **settings})

    # run the experiment
    experiment.update(1)

    # retrieve the data with data_from_database
    experiment_id = experiment.experiment_id

    # retrieve the data from emitter
    data = experiment.emitter.get_data()
    assert list(data.keys()) == [0.0, 1.0]

    db = get_experiment_database()
    # check that configuration emit was split into sub-documents
    config_cursor = db.configuration.find({'experiment_id': experiment_id})
    config_raw = list(config_cursor)
    assert len(config_raw) > 1
    # check that sim emit was split into sub-documents (2 timesteps)
    history_cursor = db.history.find({'experiment_id': experiment_id})
    history_raw = list(history_cursor)
    assert len(history_raw) > 2

    # retrieve the data directly from database
    data, experiment_config = data_from_database(experiment_id, db)

    # check values of emitted data
    processes = experiment_config['processes']
    assert len(processes) == config['number_of_processes']
    for proc in processes:
        np.testing.assert_allclose(np.array(data[1.0]['store'][proc]),
            np.array(data[0.0]['store'][proc]) + 1)

    # delete the experiment
    delete_experiment_from_database(experiment_id)


def test_query_db():
    """
    This tests the query features of the MongoDB API.
    """
    composer = ManyParametersComposite()
    composite = composer.generate()
    settings = {
        'experiment_name': 'large database experiment',
        'experiment_id': f'large_{str(uuid.uuid4())}',
        'emitter': 'database'
    }
    experiment = Engine(**{
        'processes': composite['processes'],
        'topology': composite['topology'],
        **settings})
    experiment.update(10)

    db = get_experiment_database()

    # test query
    query = [('store', 'process_0'), ('store', 'process_9')]
    data, _ = data_from_database(experiment.experiment_id, db, query)
    assert data[0.0]['store'].keys() == {'process_0', 'process_9'}

    # test func_dict
    func_dict = {
        ('store', 'process_0'): lambda x: 3,
    }
    data, _ = data_from_database(experiment.experiment_id, db, query,
                                 func_dict=func_dict)
    for emit_data in data.values():
        assert emit_data['store']['process_0'] == 3

    # test start and end time
    data, _ = data_from_database(experiment.experiment_id, db,
                                 start_time=1, end_time=5)
    assert data.keys() == {1.0, 2.0, 3.0, 4.0, 5.0}

    # test multiple cpu processes
    data_multi, _ = data_from_database(experiment.experiment_id, db, cpus=2)
    data, _ = data_from_database(experiment.experiment_id, db)
    assert data_multi == data
    delete_experiment_from_database(experiment.experiment_id, cpus=2)


def test_data_to_database():
    data = {
        '0.0': {'data': {'store': 1}, 
                'experiment_id': 'manual_insert'},
        '1.0': {'data': {'store': 2}, 
                'experiment_id': 'manual_insert'},
        '2.0': {'data': {'store': [3.5]}, 
                'experiment_id': 'manual_insert'}
    }
    config = {'experiment_id': 'manual_insert',
              'data': {'store': 1}}
    db = get_experiment_database()
    data_to_database(data, config, db)
    retrieved_data, retrieved_config = data_from_database('manual_insert', db)
    assert retrieved_data == {float(t): val['data'] for t, val in data.items()}
    assert retrieved_config == config['data']
    delete_experiment_from_database('manual_insert')
    db.configuration.delete_many({'experiment_id': 'manual_insert'})


if __name__ == '__main__':
    test_large_emits()
    test_query_db()
    test_data_to_database()



"""
===============================================
Registry of Updaters, Dividers, and Serializers
===============================================

You should interpret words and phrases that appear fully capitalized in
this document as described in :rfc:`2119`. Here is a brief summary of
the RFC:

* "MUST" indicates absolute requirements. Vivarium may not work
  correctly if you don't follow these.
* "SHOULD" indicates strong suggestions. You might have a valid reason
  for deviating from them, but be careful that you understand the
  ramifications.
* "MAY" indicates truly optional features that you can include or
  exclude as you wish.

--------
Updaters
--------

Each :term:`updater` is defined as a function whose name begins with
``update_``. Vivarium uses these functions to apply :term:`updates` to
:term:`variables`. Updater names are registered in
:py:data:`updater_registry`, which maps these names to updater functions.

Updater API
===========

An updater function SHOULD have a name that begins with ``update_``. The
function MUST accept exactly three positional arguments: the first MUST
be the current value of the variable (i.e. before applying the update),
the second MUST be the value associated with the variable in the update,
and the third MUST be either a dictionary of states from the simulation
hierarchy or None if no ``port_mapping`` key was specified in the
updater definition. The function SHOULD not accept any other parameters.
The function MUST return the updated value of the variable only.

--------
Dividers
--------

Each :term:`divider` is defined by a function that follows the API we
describe below. Vivarium uses these dividers to generate daughter cell
states from the mother cell's state. Divider names are registered in
:py:data:`divider_registry`, which maps these names to divider functions.

Divider API
===========

Each divider function SHOULD have a name prefixed with ``divide_``. The
function MUST accept a single positional argument, the value of the
variable in the mother cell. It SHOULD accept no other arguments. The
function MUST return either:

1. A :py:class:`list` with two elements: the values of the variables in
   each of the daughter cells.
2. ``None``, in which case division will be skipped for that variable.

.. note:: Dividers MAY not be deterministic and MAY not be symmetric.
    For example, a divider splitting an odd, integer-valued value may
    randomly decide which daughter cell receives the remainder.
    
-----------
Serializers
-----------

Each :term:`serializer` is defined as a class that follows the API we
describe below. Vivarium uses these serializers to convert emitted data
into a BSON-compatible format for database storage. Serializer names are
registered in :py:data:`serializer_registry`, which maps these names to
serializer subclasses. For maximum performance, register serializers
using a key equal to the string representation of its designated type
(e.g. ``str(Serializer.python_type)``).

Serializer API
==============

Serializers MUST define the following:

1. The ``python_type`` class attributes that determines what types are
   handled by the serializer
2. The :py:meth:`vivarium.core.registry.Serializer.serialize()` method
   which is called on all objects of type ``python_type``

Avoid defining custom serializers for built-in or Numpy types as these are
automatically handled by ``orjson``, the package used to serialize data.

.. note:: All dictionary keys MUST be Python strings for ``orjson`` to work.
    Numpy strings (``np.str_``) are not allowed.

If it is necessary to redefine how objects are serialized by orjson,
assign custom serializers to the stores containing objects of the affected
type(s) using the ``_serializer`` ports schema key. This can also be used
to serialize objects of the same Python type differently. To ensure that
objects serialized this way are deserialized correctly, you SHOULD consider
implementing the following as well:

1. :py:meth:`vivarium.core.registry.Serializer.can_deserialize()` to determine
   whether to call ``deserialize`` on data
2. :py:meth:`vivarium.core.registry.Serializer.deserialize()` to deserialize
   data

If it is necessary to deserialize objects of the same BSON type differently,
the corresponding serializer(s) MUST implement these 2 methods.
"""
import copy
import random
from typing import Any

import numpy as np

from vivarium.library.dict_utils import deep_merge
from vivarium.library.units import Quantity


class Registry(object):
    def __init__(self):
        """A Registry holds a collection of functions or objects."""
        self.registry = {}
        self.main_keys = []

    def register(self, key, item, alternate_keys=tuple()):
        """Add an item to the registry.

        Args:
            key: Item key.
            item: The item to add.
            alternate_keys: Additional keys under which to register the
                item. These keys will not be included in the list
                returned by ``Registry.list()``.

                This may be useful if you want to be able to look up an
                item in the registry under multiple keys.
        """
        keys = [key]
        keys.extend(alternate_keys)
        for registry_key in keys:
            if registry_key in self.registry:
                if item != self.registry[registry_key]:
                    raise Exception('registry already contains an entry for {}: {} --> {}'.format(
                        registry_key, self.registry[key], item))
            else:
                self.registry[registry_key] = item
        self.main_keys.append(key)

    def access(self, key):
        """Get an item by key from the registry."""
        return self.registry.get(key)

    def list(self):
        return list(self.main_keys)


# Initialize registries
# These are imported into module __init__.py files,
# where the functions and classes are registered upon import

#: Maps process names to :term:`process classes`
process_registry = Registry()

#: Map updater names to :term:`updater` functions
updater_registry = Registry()

#: Map divider names to :term:`divider` functions
divider_registry = Registry()

#: Map serializer names to :term:`serializer` classes
serializer_registry = Registry()

#: Map serializer names to :term:`Emitter` classes
emitter_registry = Registry()


# Updaters, Dividers, and Serializers
# These can be defined here, but are registered in the base module's __init__.py file

# Updater functions
def update_merge(current_value, new_value):
    """Merge Updater

    Returns:
        dict: The merger of ``current_value`` and ``new_value``. For any
        shared keys, the value in ``new_value`` is used.
    """
    update = current_value.copy()
    for k, v in current_value.items():
        new = new_value.get(k)
        if isinstance(new, dict):
            update[k] = deep_merge(copy.deepcopy(v), new)
        else:
            update[k] = new
    return update


def update_set(current_value, new_value):
    """Set Updater

    Returns:
        The value provided in ``new_value``.
    """
    return new_value


def update_null(current_value, new_value):
    """Null Updater

    Returns:
        The value provided in ``current_value``.
    """
    return current_value


def update_accumulate(current_value, new_value):
    """Accumulate Updater

    Returns:
        The sum of ``current_value`` and ``new_value``.
    """
    return current_value + new_value


def update_nonnegative_accumulate(current_value, new_value):
    """Non-negative Accumulate Updater

    Returns:
        The sum of ``current_value`` and ``new_value`` if positive, 0 if negative.
    """
    updated_value = current_value + new_value
    if isinstance(updated_value, np.ndarray):
        updated_value[updated_value < 0] = 0
        return updated_value
    elif updated_value >= 0:
        return updated_value
    else:
        return 0 * updated_value


def update_dictionary(current, update):
    """Dictionary Updater
    Updater that translates _add and _delete -style updates
    into operations on a dictionary.

    Expects current to be a dictionary, with no restriction on the types of objects
    stored within it, and no defaults values.
    """
    result = current

    for key, value in update.items():
        if key == "_add":
            for added_value in value:
                added_key = added_value["key"]
                added_state = added_value["state"]
                result[added_key] = added_state
        elif key == "_delete":
            for k in value:
                del result[k]
        elif key in result:
            result[key].update(value)
        else:
            raise Exception(f"Invalid dict_value_updater key: {key}")
    return result


# Divider functions
def divide_set(state):
    """Set Divider

    Returns:
        A list ``[state, state]``. No copying is performed.
    """
    return [state, state]


def divide_set_value(state, config):
    """Set Value Divider
    Args:
        'state': value
    Returns:
        A list ``[value, value]``. No copying is performed.
    """
    value = config['value']
    return [value, value]


def divide_split(state):
    """Split Divider

    Arguments:
        state: Must be an :py:class:`int`, a :py:class:`float`, or a
            :py:class:`str` of value ``Infinity``.

    Returns:
        A list, each of whose elements contains half of ``state``. If
        ``state`` is an :py:class:`int`, the remainder is placed at
        random in one of the two elements. If ``state`` is infinite, the
        return value is ``[state, state]`` (no copying is done).

    Raises:
        Exception: if ``state`` is of an unrecognized type.
    """
    if isinstance(state, (int, np.integer)):
        remainder = state % 2
        half = int(state / 2)
        if random.choice([True, False]):
            return [half + remainder, half]
        else:
            return [half, half + remainder]
    elif state == float('inf') or state == 'Infinity':
        # some concentrations are considered infinite in the environment
        # an alternative option is to not divide the local environment state
        return [state, state]
    elif isinstance(state, (float, Quantity)):
        half = state/2
        return [half, half]
    else:
        raise Exception('can not divide state {} of type {}'.format(state, type(state)))


def divide_binomial(state):
    """Binomial Divider
    """
    counts_1 = np.random.binomial(state, 0.5)
    counts_2 = state - counts_1
    return [counts_1, counts_2]


def divide_zero(state):
    """Zero Divider

    Returns:
        ``[0, 0]`` regardless of input
    """
    return [0, 0]


def divide_split_dict(state):
    """Split-Dictionary Divider

    Returns:
        A list of two dictionaries. The first dictionary stores the
        first half of the key-value pairs in ``state``, and the second
        dictionary stores the rest of the key-value pairs.

        .. note:: Since dictionaries are unordered, you should avoid
            making any assumptions about which keys will be sent to
            which daughter cell.
    """
    if state is None:
        state = {}
    d1 = dict(list(state.items())[len(state) // 2:])
    d2 = dict(list(state.items())[:len(state) // 2])
    return [d1, d2]


def assert_no_divide(state):
    """Assert that the variable is never divided

    Raises:
        AssertionError: If the variable is divided
    """
    raise AssertionError('Variable cannot be divided')


def divide_null(state):
    """Divider that causes the variable to be skipped during division.

    Returns:
        ``None`` so that no divided values are provided to the daughter
        cells. This is useful for process objects, which are handled
        separately during division.
    """
    return None



# Serializers
class Serializer:
    """Base serializer class.

    Serializers work together to convert Python objects, which may be
    collections of many different kinds of objects, into BSON-compatible
    representations. Those representations can then be deserialized to
    recover the original object.

    Serialization of Python's built-in datatypes and most Numpy types is
    handled directly by the ``orjson.dumps()`` method.
    
    The serialization routines in Serializers are compiled into a fallback
    function that is called on objects not handled by ``orjson``.

    If one wishes to modify how built-in/Numpy objects are serialized, the
    relevant stores can be assigned a custom serializer using the 
    ``_serializer`` key. The :py:meth:`vivarium.core.registry.Serializer.serialize()`
    method of that serializer will be called on the values in these stores
    before they are passed to ``orjson``.
    
    During deserialization, the ``can_deserialize`` method of each Serializer
    is used to determine whether to call the ``deserialize`` method.

    Args:
        name: Name of the serializer. Defaults to the class name.
    """
    python_type: Any = None #: Type matching is NOT exact (subclasses included)
    
    def __init__(self):
        # Register serializer under its exclusive type
        self.name = str(self.python_type) or self.__class__.__name__
    
    def serialize(self, data):
        """Controls what happens to data of the type ``python_type``
        """
        pass

    def deserialize(self, data):
        """This allows for data of the same BSON type to be deserialized
        differently (see regex matching of strings in
        :py:meth:`vivarium.core.serialize.UnitsSerializer.deserialize()`
        for an example).
        """
        pass

    def can_deserialize(self, data):
        """This tells :py:func:`vivarium.core.serialize.deserialize_value`
        whether to call ``deserialize`` on data.
        """
        pass



"""
===============
Engulf Process
===============
"""

import os

from vivarium.core.engine import pp
from vivarium.core.process import (
    Step,
)
from vivarium.core.composer import Composer, Composite
from vivarium.core.engine import Engine
from vivarium.core.directories import PROCESS_OUT_DIR
from vivarium.composites.toys import ExchangeA
from vivarium.processes.timeline import TimelineProcess


NAME = 'engulf'


class Engulf(Step):
    """ Engulf Process

    remove a compartment when the state under the 'trigger' port is set to True.
    """
    name = NAME
    defaults = {
        'agent_id': 'DEFAULT'}

    def __init__(self, parameters=None):
        super().__init__(parameters)
        self.agent_id = self.parameters['agent_id']

    def ports_schema(self):
        ''' trigger list includes ids of things to engulf '''
        return {
            'trigger': {
                '_default': [],
            },
            'inner': {
                '*': {}
            },
            'outer': {
                '*': {}
            }
        }

    def next_update(self, timestep, states):
        if states['trigger']:
            neighbors = states['trigger']
            # move neighbors from outer to inner, reset trigger
            return {
                'trigger': {
                    '_updater': 'set',
                    '_value': []},
                'outer': {
                    '_move': [{
                        # points to key in 'outer' port
                        'source': neighbor,
                        # points to which port it will be moved
                        'target': 'inner'
                    } for neighbor in neighbors]
                }
            }
        else:
            return {}


# test
class ToyAgent(Composer):
    defaults = {
        'exchange': {'uptake_rate': 0.1},
        'engulf': {
            'outer_path': ('..', '..', 'agents'),
            'inner_path': ('agents',)}}

    def generate_processes(self, config):
        return {
            'exchange': ExchangeA(config['exchange']),
            'engulf': Engulf(config['engulf']),
            'expel': Engulf(config['engulf'])}

    def generate_topology(self, config):
        return {
            'exchange': {
                'external': config['exchange']['external_path'],
                'internal': config['exchange']['internal_path']},
            'engulf': {
                'trigger': ('engulf-trigger',),
                'inner': config['engulf']['inner_path'],
                'outer': config['engulf']['outer_path']},
            'expel': {
                'trigger': {
                    '_path': ('expel-trigger',)},
                'inner': config['engulf']['outer_path'],
                'outer': config['engulf']['inner_path']}}


def test_engulf(return_value=False):
    num_agents = 3
    agent_ids = [
        str(agent_id + 1)
        for agent_id in range(num_agents)]

    # initial state
    initial_state = {
        'concentrations': {'A': 10.0},
        'agents': {
            agent_id: {
                'concentrations': {'A': float(int(agent_id))},
                'trigger': []}
            for agent_id in agent_ids}}

    # timeline triggers engulf for agent_1
    time_total = 10
    timeline = [
        (0, {('agents', agent_ids[0], 'trigger'): []}),
        (3, {('agents', agent_ids[2], 'engulf-trigger'): [agent_ids[1]]}),
        (5, {('agents', agent_ids[0], 'engulf-trigger'): [agent_ids[2]]}),
        (8, {('agents', agent_ids[0], 'agents', agent_ids[2], 'expel-trigger'): [agent_ids[1]]}),
        (time_total, {})]

    # declare the hierarchy
    timeline = TimelineProcess({'timeline': timeline})
    agent = ToyAgent({
        'exchange': {
            'internal_path': ('concentrations',),
            'external_path': ('..', '..', 'concentrations')},
        'engulf': {
            'inner_path': ('agents',),
            'outer_path': ('..', '..', 'agents')}})

    composite = Composite()
    for agent_id in agent_ids:
        composite.merge(agent.generate(path=('agents', agent_id)))
    composite.merge(
        processes={'timeline': timeline},
        topology={'timeline': {
            'global': ('global',),
            'agents': ('agents',)}},
        state=initial_state
    )

    # configure experiment
    experiment = Engine(
        composite=composite,
    )

    # run simulation
    experiment.update(time_total)
    output = experiment.emitter.get_data()
    experiment.end()  # end required for parallel processes

    # assert the engulfing/expelling saga
    assert [*output[0.0]['agents'].keys()] == agent_ids
    assert [*output[4.0]['agents'].keys()] == ['1', '3']
    assert [*output[6.0]['agents']['1']['agents'].keys()] == ['3']
    assert [*output[10.0]['agents'].keys()] == ['1']
    assert [*output[10.0]['agents']['1']['agents'].keys()] == ['3', '2']

    if return_value:
        return output


def run_engulf():
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    os.makedirs(out_dir, exist_ok=True)
    output = test_engulf(return_value=True)
    pp(output)


if __name__ == '__main__':
    run_engulf()



import random
import logging as log
from typing import Optional, Union, Dict, Any, cast, List

from vivarium.composites.toys import (
    PoQo, Sine, ToyDivider, ToyTransport, ToyEnvironment, ToyProcess,
    Proton, Electron, MoveProcess)
from vivarium.core.composer import Composer, Composite
from vivarium.core.engine import Engine, pf, pp, _StepGraph
from vivarium.core.process import Process, Step, Deriver
from vivarium.core.store import Store, hierarchy_depth
from vivarium.core.types import (
    Schema, State, Update, Topology, Steps, Processes)
from vivarium.library.units import units
from vivarium.library.wrappers import make_logging_process
from vivarium.core.control import run_library_cli


def _make_proton(
        parallel: bool = False
) -> Dict[str, Any]:
    processes = {
        'proton': Proton({'_parallel': parallel}),
        'electrons': {
            'a': {
                'electron': Electron({'_parallel': parallel})},
            'b': {
                'electron': Electron()}}}

    spin_path = ('internal', 'spin')
    radius_path = ('structure', 'radius')

    topology = {
        'proton': {
            'radius': radius_path,
            'quarks': ('internal', 'quarks'),
            'electrons': {
                '_path': ('electrons',),
                '*': {
                    'orbital': ('shell', 'orbital'),
                    'spin': spin_path}}},
        'electrons': {
            'a': {
                'electron': {
                    'spin': spin_path,
                    'proton': {
                        '_path': ('..', '..'),
                        'radius': radius_path}}},
            'b': {
                'electron': {
                    'spin': spin_path,
                    'proton': {
                        '_path': ('..', '..'),
                        'radius': radius_path}}}}}

    initial_state = {
        'structure': {
            'radius': 0.7},
        'internal': {
            'quarks': {
                'x': {
                    'color': 'green',
                    'spin': 'up'},
                'y': {
                    'color': 'red',
                    'spin': 'up'},
                'z': {
                    'color': 'blue',
                    'spin': 'down'}}}}

    return {
        'processes': processes,
        'topology': topology,
        'initial_state': initial_state}


def test_recursive_store() -> None:
    environment_config = {
        'environment': {
            'temperature': {
                '_default': 0.0,
                '_updater': 'accumulate'},
            'fields': {
                (0, 1): {
                    'enzymeX': {
                        '_default': 0.0,
                        '_updater': 'set'},
                    'enzymeY': {
                        '_default': 0.0,
                        '_updater': 'set'}},
                (0, 2): {
                    'enzymeX': {
                        '_default': 0.0,
                        '_updater': 'set'},
                    'enzymeY': {
                        '_default': 0.0,
                        '_updater': 'set'}}},
            'agents': {
                '1': {
                    'location': {
                        '_default': (0, 0),
                        '_updater': 'set'},
                    'boundary': {
                        'external': {
                            '_default': 0.0,
                            '_updater': 'set'},
                        'internal': {
                            '_default': 0.0,
                            '_updater': 'set'}},
                    'transcripts': {
                        'flhDC': {
                            '_default': 0,
                            '_updater': 'accumulate'},
                        'fliA': {
                            '_default': 0,
                            '_updater': 'accumulate'}},
                    'proteins': {
                        'ribosome': {
                            '_default': 0,
                            '_updater': 'set'},
                        'flagella': {
                            '_default': 0,
                            '_updater': 'accumulate'}}},
                '2': {
                    'location': {
                        '_default': (0, 0),
                        '_updater': 'set'},
                    'boundary': {
                        'external': {
                            '_default': 0.0,
                            '_updater': 'set'},
                        'internal': {
                            '_default': 0.0,
                            '_updater': 'set'}},
                    'transcripts': {
                        'flhDC': {
                            '_default': 0,
                            '_updater': 'accumulate'},
                        'fliA': {
                            '_default': 0,
                            '_updater': 'accumulate'}},
                    'proteins': {
                        'ribosome': {
                            '_default': 0,
                            '_updater': 'set'},
                        'flagella': {
                            '_default': 0,
                            '_updater': 'accumulate'}}}}}}

    state = Store(environment_config)
    state.apply_update({})
    state.state_for(['environment'], ['temperature'])


def test_topology_ports() -> None:
    proton = _make_proton()

    experiment = Engine(**proton)

    log.debug(pf(experiment.state.get_config(True)))

    experiment.update(10.0)

    log.debug(pf(experiment.state.get_config(True)))
    log.debug(pf(experiment.state.divide_value()))


def test_timescales() -> None:
    class Slow(Process):
        name = 'slow'
        defaults = {'timestep': 3.0}

        def __init__(self, config: Optional[dict] = None) -> None:
            super().__init__(config)

        def ports_schema(self) -> Schema:
            return {
                'state': {
                    'base': {
                        '_default': 1.0}}}

        def next_update(
                self,
                timestep: Union[float, int],
                states: State) -> Update:
            base = states['state']['base']
            next_base = timestep * base * 0.1

            return {
                'state': {'base': next_base}}

    class Fast(Process):
        name = 'fast'
        defaults = {'timestep': 0.3}

        def __init__(self, config: Optional[dict] = None) -> None:
            super().__init__(config)

        def ports_schema(self) -> Schema:
            return {
                'state': {
                    'base': {
                        '_default': 1.0},
                    'motion': {
                        '_default': 0.0}}}

        def next_update(
                self,
                timestep: Union[float, int],
                states: State) -> Update:
            base = states['state']['base']
            motion = timestep * base * 0.001

            return {
                'state': {'motion': motion}}

    processes = {
        'slow': Slow(),
        'fast': Fast()}

    states = {
        'state': {
            'base': 1.0,
            'motion': 0.0}}

    topology: Topology = {
        'slow': {'state': ('state',)},
        'fast': {'state': ('state',)}}

    emitter = {'type': 'null'}
    experiment = Engine(
        processes=processes,
        topology=topology,
        emitter=emitter,
        initial_state=states)

    experiment.update(10.0)


def test_2_store_1_port() -> None:
    """
    Split one port of a processes into two stores
    """

    class OnePort(Process):
        name = 'one_port'

        def ports_schema(self) -> Schema:
            return {
                'A': {
                    'a': {
                        '_default': 0,
                        '_emit': True},
                    'b': {
                        '_default': 0,
                        '_emit': True}
                }
            }

        def next_update(
                self,
                timestep: Union[float, int],
                states: State) -> Update:
            return {
                'A': {
                    'a': 1,
                    'b': 2}}

    class SplitPort(Composer):
        """splits OnePort's ports into two stores"""
        name = 'split_port_composer'

        def generate_processes(
                self, config: Optional[dict]) -> Dict[str, Any]:
            return {
                'one_port': OnePort({})}

        def generate_topology(self, config: Optional[dict]) -> Topology:
            return {
                'one_port': {
                    'A': {
                        'a': ('internal', 'a',),
                        'b': ('external', 'a',)
                    }
                }}

    # run experiment
    split_port = SplitPort({})
    network = split_port.generate()
    exp = Engine(**{
        'processes': network['processes'],
        'topology': network['topology']})

    exp.update(2)
    output = exp.emitter.get_timeseries()
    expected_output = {
        'external': {'a': [0, 2, 4]},
        'internal': {'a': [0, 1, 2]},
        'time': [0.0, 1.0, 2.0]}
    assert output == expected_output


class MultiPort(Process):
    name = 'multi_port'

    def ports_schema(self) -> Schema:
        return {
            'A': {
                'a': {
                    '_default': 0,
                    '_emit': True}},
            'B': {
                'a': {
                    '_default': 0,
                    '_emit': True}},
            'C': {
                'a': {
                    '_default': 0,
                    '_emit': True}}}

    def next_update(
            self,
            timestep: Union[float, int],
            states: State) -> Update:
        return {
            'A': {'a': 1},
            'B': {'a': 1},
            'C': {'a': 1}}


class MergePort(Composer):
    """combines both of MultiPort's ports into one store"""
    name = 'multi_port_composer'

    def generate_processes(
            self, config: Optional[dict]) -> Dict[str, Any]:
        return {
            'multi_port': MultiPort({})}

    def generate_topology(self, config: Optional[dict]) -> Topology:
        return {
            'multi_port': {
                'A': ('aaa',),
                'B': ('aaa',),
                'C': ('aaa',)}}


def test_multi_port_merge() -> None:
    # run experiment
    merge_port = MergePort({})
    network = merge_port.generate()
    exp = Engine(**{
        'processes': network['processes'],
        'topology': network['topology']})

    exp.update(2)
    output = exp.emitter.get_timeseries()
    expected_output = {
        'aaa': {'a': [0, 3, 6]},
        'time': [0.0, 1.0, 2.0]}

    assert output == expected_output


def test_emit_config() -> None:
    # test alternate emit options
    merge_port = MergePort({})
    network = merge_port.generate()
    exp1 = Engine(
        processes=network['processes'],
        topology=network['topology'],
        emit_topology=False,
        emit_processes=True,
        emit_config=True,
        progress_bar=True,
        emit_step=2,
    )

    exp1.update(10)


def test_complex_topology() -> None:
    # make the experiment
    outer_path = ('universe', 'agent')
    pq = PoQo({})
    pq_composite = pq.generate(path=outer_path)
    pq_composite.pop('_schema')
    experiment = Engine(composite=pq_composite)

    # get the initial state
    initial_state = experiment.state.get_value()
    print('time 0:')
    pp(initial_state)

    # simulate for 1 second
    experiment.update(1)

    next_state = experiment.state.get_value()
    print('time 1:')
    pp(next_state)

    # pull out the agent state
    initial_agent_state = initial_state['universe']['agent']
    agent_state = next_state['universe']['agent']

    assert agent_state['aaa']['a1'] == initial_agent_state['aaa']['a1'] + 1
    assert agent_state['aaa']['x'] == initial_agent_state['aaa']['x'] - 9
    assert agent_state['ccc']['a3'] == initial_agent_state['ccc']['a3'] + 1


def test_parallel() -> None:
    proton = _make_proton(parallel=True)
    experiment = Engine(**proton)

    log.debug(pf(experiment.state.get_config(True)))

    experiment.update(10.0)

    log.debug(pf(experiment.state.get_config(True)))
    log.debug(pf(experiment.state.divide_value()))

    experiment.end()


def test_depth() -> None:
    nested = {
        'A': {
            'AA': 5,
            'AB': {
                'ABC': 11}},
        'B': {
            'BA': 6}}

    dissected = hierarchy_depth(nested)
    assert len(dissected) == 3
    assert dissected[('A', 'AB', 'ABC')] == 11


def test_sine() -> None:
    sine = Sine()
    print(sine.next_update(0.25 / 440.0, {
        'frequency': 440.0,
        'amplitude': 0.1,
        'phase': 1.5}))


def test_units() -> None:
    class UnitsMicrometer(Process):
        name = 'units_micrometer'

        def ports_schema(self) -> Schema:
            return {
                'A': {
                    'a': {
                        '_default': 0 * units.um,
                        '_emit': True},
                    'b': {
                        '_default': 'string b',
                        '_emit': True,
                    },
                    'c': {
                        '_default': 0,
                        '_emit': True,
                    }
                }
            }

        def next_update(
                self, timestep: Union[float, int], states: State) -> Update:
            return {
                'A': {
                    'a': 1 * units.um,
                    'c': 1,
                }}

    class UnitsMillimeter(Process):
        name = 'units_millimeter'

        def ports_schema(self) -> Schema:
            return {
                'A': {
                    'a': {
                        # '_default': 0 * units.mm,
                        '_emit': True}}}

        def next_update(
                self, timestep: Union[float, int], states: State) -> Update:
            return {
                'A': {'a': 1 * units.mm}}

    class MultiUnits(Composer):
        name = 'multi_units_composer'

        def generate_processes(
                self,
                config: Optional[dict]) -> Dict[str, Any]:
            return {
                'units_micrometer':
                    UnitsMicrometer({}),
                'units_millimeter':
                    UnitsMillimeter({})}

        def generate_topology(self, config: Optional[dict]) -> Topology:
            return {
                'units_micrometer': {
                    'A': ('aaa',)},
                'units_millimeter': {
                    'A': ('aaa',)}}

    # run experiment
    multi_unit = MultiUnits({})
    network = multi_unit.generate()
    exp = Engine(**{
        'processes': network['processes'],
        'topology': network['topology']})

    exp.update(5)
    timeseries = exp.emitter.get_timeseries()
    print('TIMESERIES')
    pp(timeseries)

    data = exp.emitter.get_data()
    print('DATA')
    pp(data)

    data_deserialized = exp.emitter.get_data_deserialized()
    print('DESERIALIZED')
    pp(data_deserialized)

    data_unitless = exp.emitter.get_data_unitless()
    print('UNITLESS')
    pp(data_unitless)

    query = [('aaa', 'a'), ('aaa', 'c')]
    query_data = exp.emitter.get_data(query)
    print('QUERY DATA')
    pp(query_data)


def test_custom_divider() -> None:
    """ToyDividerProcess has a custom `split_divider`"""
    agent_id = '1'
    composer = ToyDivider({
        'agent_id': agent_id,
        'divider': {
            'x_division_threshold': 3,
        }
    })
    composite = composer.generate(path=('agents', agent_id))

    experiment = Engine(
        processes=composite.processes,
        steps=composite.steps,
        flow=composite.flow,
        topology=composite.topology,
    )

    experiment.update(8)
    data = experiment.emitter.get_data()

    expected_data = {
        0.0: {'agents': {'1': {'variable': {'x': 0, '2x': 0}}}},
        1.0: {'agents': {'1': {'variable': {'x': 1, '2x': 2}}}},
        2.0: {'agents': {'1': {'variable': {'x': 2, '2x': 4}}}},
        3.0: {'agents': {'1': {'variable': {'x': 3, '2x': 6}}}},
        4.0: {'agents': {'1': {'variable': {'x': 4, '2x': 8}}}},
        5.0: {'agents': {'10': {'variable': {'x': 2, '2x': 4}},
                         '11': {'variable': {'x': 2, '2x': 4}}}},
        6.0: {'agents': {'10': {'variable': {'x': 3, '2x': 6}},
                         '11': {'variable': {'x': 3, '2x': 6}}}},
        7.0: {'agents': {'10': {'variable': {'x': 4, '2x': 8}},
                         '11': {'variable': {'x': 4, '2x': 8}}}},
        8.0: {'agents': {'100': {'variable': {'x': 2, '2x': 4}},
                         '101': {'variable': {'x': 2, '2x': 4}},
                         '110': {'variable': {'x': 2, '2x': 4}},
                         '111': {'variable': {'x': 2, '2x': 4}}}}
    }
    assert data == expected_data


class TestStepGraph:

    @staticmethod
    def test_step_graph_execution_layers() -> None:
        tg = _StepGraph()
        tg.add(('a',), [])
        tg.add(('b',), [])
        tg.add(('c',), [('a',), ('b',)])
        tg.add(('d',), [('c',)])

        layers = list(tg.get_execution_layers())
        expected_layers = [
            [('a',), ('b',)],
            [('c',)],
            [('d',)],
        ]
        assert layers == expected_layers

    @staticmethod
    def test_step_graph_sequential() -> None:
        tg = _StepGraph()
        tg.add(('a',), [])
        tg.add(('b',), [])
        tg.add_sequential(('c',))

        layers = list(tg.get_execution_layers())
        expected_layers = [
            [('c',)],
            [('a',), ('b',)],
        ]
        assert layers == expected_layers

    @staticmethod
    def test_step_graph_removal() -> None:
        tg = _StepGraph()
        tg.add(('a',), [])
        tg.add(('b',), [('a',)])
        tg.add(('c',), [('b',)])

        tg.remove(('a',))
        layers = list(tg.get_execution_layers())

        assert not layers


def test_runtime_order() -> None:
    class RuntimeOrderProcess(Process):

        def ports_schema(self) -> Schema:
            return {
                'store': {
                    'var': {
                        '_default': 0
                    }
                }
            }

        def next_update(self, timestep: float, states: State) -> Update:
            _ = states
            self.parameters['execution_log'].append(self.name)
            return {}

    class RuntimeOrderStep(Step):

        def ports_schema(self) -> Schema:
            return {
                'store': {
                    'var': {
                        '_default': 0
                    }
                }
            }

        def next_update(self, timestep: float, states: State) -> Update:
            _ = states
            self.parameters['execution_log'].append(self.name)
            return {}

    class RuntimeOrderDeriver(Deriver):

        def ports_schema(self) -> Schema:
            return {
                'store': {
                    'var': {
                        '_default': 0
                    }
                }
            }

        def next_update(self, timestep: float, states: State) -> Update:
            _ = states
            self.parameters['execution_log'].append(self.name)
            return {}

    class RuntimeOrderComposer(Composer):

        def generate_processes(
                self, config: Optional[dict]) -> Dict[str, Any]:
            config = cast(dict, config or {})
            proc1 = RuntimeOrderProcess({
                'name': 'process1',
                'time_step': 1,
                'execution_log': config['execution_log'],
            })
            proc2 = RuntimeOrderProcess({
                'name': 'process2',
                'time_step': 2,
                'execution_log': config['execution_log'],
            })
            deriver = RuntimeOrderDeriver({
                'name': 'deriver',
                'execution_log': config['execution_log'],
            })
            return {
                'p1': proc1,
                'p2': proc2,
                'd': deriver,
            }

        def generate_steps(self, config: Optional[dict]) -> Steps:
            config = config or {}
            step1 = RuntimeOrderStep({
                'name': 'step1',
                'execution_log': config['execution_log'],
            })
            step2 = RuntimeOrderStep({
                'name': 'step2',
                'execution_log': config['execution_log'],
            })
            step3 = RuntimeOrderStep({
                'name': 'step3',
                'execution_log': config['execution_log'],
            })
            return {
                's1': step1,
                's2': step2,
                's3': step3,
            }

        def generate_flow(self, config: Optional[dict]) -> Steps:
            config = config or {}
            return {
                's1': [],
                's2': [('s1',)],
                's3': [('s1',)],
            }

        def generate_topology(self, config: Optional[dict]) -> Topology:
            return {
                'p1': {
                    'store': ('store',),
                },
                'p2': {
                    'store': ('store',),
                },
                'd': {
                    'store': ('store',),
                },
                's1': {
                    'store': ('store',),
                },
                's2': {
                    'store': ('store',),
                },
                's3': {
                    'store': ('store',),
                },
            }

    execution_log: List[str] = []
    composer = RuntimeOrderComposer()
    composite = composer.generate({'execution_log': execution_log})
    experiment = Engine(
        processes=composite.processes,
        steps=composite.steps,
        flow=composite.flow,
        topology=composite.topology,
    )
    experiment.update(4)
    expected_log = [
        ('deriver', 'step1'),
        {'step2', 'step3'},
        {'process1', 'process2'},
        ('deriver', 'step1'),
        {'step2', 'step3'},
        {'process1'},
        ('deriver', 'step1'),
        {'step2', 'step3'},
        {'process1', 'process2'},
        ('deriver', 'step1'),
        {'step2', 'step3'},
        {'process1'},
        ('deriver', 'step1'),
        {'step2', 'step3'},
    ]
    for expected_group in expected_log:
        num = len(expected_group)
        group = execution_log[0:num]
        execution_log = execution_log[num:]
        if isinstance(expected_group, tuple):
            assert tuple(group) == expected_group
        elif isinstance(expected_group, set):
            assert set(group) == expected_group


def get_toy_transport_in_env_composite(
        agent_id: Any = '0',
        transport_config: Optional[dict] = None,
        environment_config: Optional[dict] = None,
) -> Composite:
    processes = {
        'agents': {
            agent_id: {
                'transport': ToyTransport(transport_config)}},
        'environment': ToyEnvironment(environment_config)}
    topology = {
        'environment': {
            'agents': {
                '_path': ('agents',),
                '*': {
                    'external': ('external', 'GLC')}}},
        'agents': {
            agent_id: {
                'transport': {
                    'internal': ('internal',),
                    'external': ('external',)}}}}
    return Composite({
        'processes': processes,
        'topology': topology,
    })


def test_glob_schema() -> None:
    composite = get_toy_transport_in_env_composite()
    experiment = Engine(
        processes=composite.processes,
        topology=composite.topology)
    experiment.update(10)

    # declare processes in reverse order
    processes_reverse = {
        'environment': ToyEnvironment(),
        'agents': {'0': {'transport': ToyTransport()}}}

    experiment_reverse = Engine(
        processes=processes_reverse,
        topology=composite.topology)
    experiment_reverse.update(10)


def get_env_view_composite(agent_id: Any = '1') -> Composite:
    agent_composer = ToyDivider({
        'agent_id': agent_id,
        'divider': {
            'x_division_threshold': 3,
        }
    })
    composite = agent_composer.generate(path=('agents', agent_id))

    environment_process = {
        'environment': make_logging_process(ToyEnvironment)()
    }
    environment_topology: Topology = {
        'environment': {
            'agents': {
                '_path': ('agents',),
                '*': {
                    'external': ('external', 'GLC')
                }
            },
            'log_update': ('log_update',),
        }
    }

    # combine the environment and agent
    composite.merge(
        processes=environment_process,
        topology=environment_topology,
    )
    return composite


def test_environment_view_with_division() -> None:
    composite = get_env_view_composite()
    experiment = Engine(
        processes=composite.processes,
        topology=composite.topology)
    experiment.update(10)
    data = experiment.emitter.get_data()

    # confirm that the environment sees the new agents.
    once_different = False
    for state in data.values():
        agent_ids = set(state['agents'].keys())
        env_agents = set(state['log_update'].get('agents', {}).keys())
        if env_agents != agent_ids:
            if not once_different:
                once_different = True
            else:
                # the values have been different for more than one update
                raise ValueError(
                    f'environment sees {env_agents} instead of {agent_ids}')
        else:
            once_different = False


class AddDelete(Process):
    def ports_schema(self) -> Schema:
        return {
            'sub_stores': {
                '*': {
                    '_default': 0,
                    '_emit': True
                }
            },
            'expected': {
                '_default': [],
                '_updater': 'set',
            }
        }

    def next_update(
            self, timestep: Union[float, int], states: State) -> Update:
        sub_stores = set(states['sub_stores'].keys())
        expected = set(states['expected'])
        assert sub_stores == expected, "stores don't match expected"

        # delete current stores, and add the same number of stores
        sub_stores_update: dict = {
            '_delete': [],
            '_add': []}
        new_sub_stores = []
        for store_id in sub_stores:
            sub_stores_update['_delete'].append(store_id)
            new_id = str(random.randint(0, 2 ** 63))
            new_sub_stores.append(new_id)
            new_store = {'key': new_id, 'state': 1}
            sub_stores_update['_add'].append(new_store)

        return {
            'sub_stores': sub_stores_update,
            'expected': new_sub_stores,
        }


def test_add_delete() -> None:
    process = AddDelete()
    topology = {
        'sub_stores': ('sub_stores',),
        'expected': ('expected',)}

    # initial state
    n_initial = 10
    initial_substores = [
        str(random.randint(0, 2**63)) for _ in range(n_initial)]
    initial_state = {
        'sub_stores': {
            sub_store: 1
            for sub_store in initial_substores
        },
        'expected': initial_substores,
    }

    experiment = Engine(
        processes={'process': process},
        topology={'process': topology},
        initial_state=initial_state,
    )
    experiment.update(10)

    # assert that no overlapping sub store between time steps.
    # All sub stores should get deleted, and new sub stores added
    data = experiment.emitter.get_data()
    times = list(data.keys())
    n_times = len(times)
    for t_index in range(n_times):
        if t_index < n_times - 1:
            current_time = times[t_index]
            next_time = times[t_index + 1]
            current_ids = set(data[current_time]['sub_stores'].keys())
            next_ids = set(data[next_time]['sub_stores'].keys())
            assert len(set(current_ids).intersection(set(next_ids))) == 0


def test_hyperdivision(profile: bool = True) -> None:
    total_time = 10
    n_agents = 100
    division_thresholds = [3, 4, 5, 6, 7]  # what values of x triggers division?

    # initialize agent composer
    agent_composer = ToyDivider()

    # make the composite
    composite = Composite()
    agent_ids = [str(agent_idx) for agent_idx in range(n_agents)]
    for agent_id in agent_ids:
        divider_config = {
            'divider': {
                    'x_division_threshold': random.choice(division_thresholds),
                }}
        agent_composite = agent_composer.generate(
            config={
                'agent_id': agent_id,
                **divider_config,
            },
            path=('agents', agent_id))
        composite.merge(agent_composite)

    # add an environment
    environment_process: Processes = {'environment': ToyEnvironment()}
    environment_topology: Topology = {
        'environment': {
            'agents': {
                '_path': ('agents',),
                '*': {
                    'external': ('external', 'GLC')
                }
            },
        }
    }

    # combine the environment and agent
    composite.merge(
        processes=environment_process,
        topology=environment_topology,
    )

    # make the sim, run the sim, retrieve the data
    experiment = Engine(
        processes=composite.processes,
        steps=composite.steps,
        flow=composite.flow,
        topology=composite.topology,
        profile=profile,
    )
    experiment.update(total_time)
    experiment.end()
    data = experiment.emitter.get_data()

    print(f"n agents initial: {n_agents}")
    print(f"n agents final: {len(data[total_time]['agents'].keys())}")
    assert len(data[total_time]['agents'].keys()) > n_agents

    if profile:
        stats = experiment.stats
        stats.strip_dirs().sort_stats(  # type: ignore
            'cumulative', 'cumtime').print_stats(20)

        # make sure view_values is fast
        stats_view_values = stats.get_print_list(  # type: ignore
            ('view_values',))[1]
        view_values_times = stats.stats[  # type: ignore
            stats_view_values[0]][3]
        total_runtime = stats.total_tt  # type: ignore
        assert view_values_times < 0.1 * total_runtime

def test_output_port() -> None:
    a_default = 1
    b_default = 2

    class InputOutput(Process):
        def ports_schema(self) -> Schema:
            return {
                'input': {
                    'A': {
                        '_default': a_default,
                        '_emit': True,
                    }
                },
                'output': {
                    '_output': True,
                    'B': {
                        '_default': b_default,
                        '_emit': True,
                    }
                }
            }
        def next_update(
                self,
                timestep: Union[float, int],
                states: State) -> Update:
            assert not states['output'], 'outputs should be masked'
            return {}

    total_time = 10
    composite = InputOutput().generate()
    sim = Engine(
        processes=composite['processes'],
        topology=composite['topology']
    )
    sim.update(total_time)
    data = sim.emitter.get_timeseries()
    assert data['input']['A'] == [a_default for _ in range(total_time + 1)]
    assert data['output']['B'] == [b_default for _ in range(total_time + 1)]


def test_engine_run_for() -> None:
    total_time = 10.0
    time_interval = 1.0
    timestep1 = 0.75
    timestep2 = 1.25

    topo = {
        'external': ('external',),
        'internal': ('internal',),
    }
    composite = Composite({
        'processes': {
            'process1': ToyTransport({'time_step': timestep1}),
            'process2': ToyTransport({'time_step': timestep2}),
        },
        'topology': {
            'process1': topo,
            'process2': topo,
        }
    })
    initial_state = {
        'external': {'GLC': 100},
    }

    sim = Engine(
        processes=composite.processes,
        topology=composite.topology,
        initial_state=initial_state)

    time = 0.0
    while sim.global_time < total_time:
        if time + time_interval >= total_time:
            # If this is the last iteration of the simulation loop,
            # force processes to complete.
            sim.run_for(time_interval, force_complete=True)
        else:
            sim.run_for(time_interval)
        time += time_interval
        assert sim.global_time == time

        # check that the front has advanced correctly
        front = sim.front
        for path, advance in front.items():
            expected_time = 0.0
            if time >= total_time:
                expected_time = total_time
            elif path[0] == 'process1':
                expected_time = int(time / timestep1) * timestep1
            elif path[0] == 'process2':
                expected_time = int(time / timestep2) * timestep2
            assert advance['time'] == expected_time, \
                f"front time {advance['time']} " \
                f"is not expected {expected_time}"

    final_time = time
    assert sim.global_time == final_time

    front = sim.front
    for path, advance in front.items():
        assert advance['time'] == sim.global_time, \
            f"process at path {path} did not complete"


def test_set_branch_emit() -> None:
    run_time = 5
    # get the composer
    composer = PoQo({})

    # turn on emits
    composite = composer.generate()
    exp = Engine(
        composite=composite,
        store_schema={'_emit': True}
    )
    exp.update(run_time)
    data = exp.emitter.get_data()
    assert data[run_time]['bbb'] != {}, 'this emit should be on'
    print(pf(data))

    # turn off emits
    composite = composer.generate()
    exp = Engine(
        composite=composite,
        store_schema={'_emit': False}
    )
    exp.update(run_time)
    data = exp.emitter.get_data()
    assert data[run_time]['bbb'] == {}, 'this emit should be off'
    print(pf(data))

    # selectively turn on emits
    composite = composer.generate()
    exp = Engine(
        composite=composite,
        store_schema={'bbb': {'e2': {'_emit': True}}},
    )
    exp.update(run_time)
    data = exp.emitter.get_data()
    assert data[run_time]['bbb']['e2'] != {}, 'this emit should be on'
    assert data[run_time]['ccc'] == {}, 'this emit should be off'
    print(pf(data))


def test_add_new_state() -> None:
    agent_id = '1'
    composite = get_toy_transport_in_env_composite(agent_id=agent_id)
    new_schema = {
        'agents': {
            agent_id: {
                'extra': {'_emit': True, '_value': 1.0}}}}
    experiment = Engine(
        processes=composite.processes,
        topology=composite.topology,
        store_schema=new_schema,
    )

    assert experiment.state['agents', agent_id, 'extra'].get_value() == 1.0

    total_time = 5
    experiment.update(total_time)
    timeseries = experiment.emitter.get_timeseries()
    assert len(timeseries['agents'][agent_id]['extra']) == total_time + 1


def test_floating_point_timesteps() -> None:
    agent_id = '1'
    transport_timestep = 0.1
    composite = get_toy_transport_in_env_composite(
        agent_id=agent_id,
        transport_config={'time_step': transport_timestep}
    )
    sim = Engine(
        processes=composite.processes,
        topology=composite.topology,
        global_time_precision=5
    )
    sim.update(1)
    data = sim.emitter.get_data()
    print(pf(data))
    for t in data.keys():
        ntimes = round(t / transport_timestep, 5)
        expected_t = round(ntimes * transport_timestep, 5)
        assert t == expected_t, f'bad timestep {t}'


def test_move_update() -> None:
    agents_in_1 = ['a1', 'a2']
    agents_in_2 = ['a3', 'a4']
    agents_in_3 = ['a5', 'a6']

    composite = Composite({
        'processes': {
            'move': MoveProcess({'timestep': 2}),
            'store1': {
                agent_id: {
                    'toy': ToyProcess({'timestep': 1})
                } for agent_id in agents_in_1},
        },
        'topology': {
            'move': {
                '1': ('store1',),
                '2': ('store2',),
                '3': ('store3',),
            },
            'store1': {
                agent_id: {
                    'toy': {
                        'port1': ('store_A',),
                        'port2': ('store_B',)
                    }
                } for agent_id in agents_in_1},
        }})

    state = {
        'store1': {
            agent_id: {
                'store_A': {'var_a': 0.1, 'var_b': 0.2},
                'store_B': {'var_a': 0.3, 'var_b': 0.4},
            } for agent_id in agents_in_1
        },
        'store2': {
            agent_id: {} for agent_id in agents_in_2
        },
        'store3': {
            agent_id: {} for agent_id in agents_in_3
        },
    }

    sim = Engine(
        processes=composite.processes,
        topology=composite.topology,
        initial_state=state,
    )

    sim.update(4)
    data = sim.emitter.get_data()

    # it should move at t=2, t=4
    assert list(data[4]['store1'].keys()) == agents_in_2
    assert list(data[4]['store2'].keys()) == agents_in_3
    assert list(data[4]['store3'].keys()) == agents_in_1
    assert 'var_a' in data[4]['store3'][agents_in_1[0]]['store_A'], \
        'the agent initialized in store1 should have ' \
        'carried its internal states generated by ToyProcess'



engine_tests = {
    '0': test_recursive_store,
    '1': test_topology_ports,
    '2': test_timescales,
    '3': test_2_store_1_port,
    '4': test_multi_port_merge,
    '5': test_emit_config,
    '6': test_complex_topology,
    '7': test_parallel,
    '8': test_depth,
    '9': test_sine,
    '10': test_units,
    '11': test_custom_divider,
    '12': test_runtime_order,
    '13': test_glob_schema,
    '14': test_environment_view_with_division,
    '15': test_add_delete,
    '16': test_hyperdivision,
    '17': test_output_port,
    '18': test_engine_run_for,
    '19': test_set_branch_emit,
    '20': test_add_new_state,
    '21': test_floating_point_timesteps,
    '22': test_move_update,
}


# python vivarium/experiments/engine_tests.py -n [test number]
if __name__ == '__main__':
    run_library_cli(engine_tests)



import numpy as np

from vivarium.core.engine import Engine
from vivarium.core.composer import Composer
from vivarium.library.units import units
from vivarium.processes.divide_condition import DivideCondition
from vivarium.processes.meta_division import MetaDivision
from vivarium.processes.growth_rate import GrowthRate


TIMESTEP = 10


class AgentDivision(Composer):
    defaults = {
        'time_step': TIMESTEP,
        'Tl': {},
        'concs': {
            'molecular_weights': {
                'C': 1e8 * units.g / units.mol}},
        'growth': {
            'time_step': 1,
            'default_growth_rate': 0.0005,
            'default_growth_noise': 0.001,
            'variables': ['volume']},
        'agent_id': np.random.randint(0, 100),
        'divide_condition': {
            'threshold': 2.5 * units.fL},
        'agents_path': ('..', '..', 'agents',),
        'boundary_path': ('boundary',),
        'daughter_path': tuple(),
    }

    def generate_processes(self, config) -> dict:
        division_config = {
            'daughter_path': config['daughter_path'],
            'agent_id': config['agent_id'],
            'composer': self}
        time_step_config = {'time_step': config['time_step']}
        return {
            'growth': GrowthRate({**config['growth'], **time_step_config}),
            'divide_condition': DivideCondition(config['divide_condition']),
            'division': MetaDivision(division_config)}

    def generate_topology(self, config) -> dict:
        boundary_path = config['boundary_path']
        agents_path = config['agents_path']
        return {
            'growth': {
                'variables': boundary_path,
                'rates': ('rates',)},
            'divide_condition': {
                'variable': boundary_path + ('volume',),
                'divide': boundary_path + ('divide',)},
            'division': {
                'global': boundary_path,
                'agents': agents_path}}


def test_hierarchy_update() -> None:
    # configure hierarchy
    agent_id = '0'

    # initial state
    initial_state = {
        'agents': {
            agent_id: {
                'boundary': {'volume': 1.2 * units.fL},
                'DNA': {'G': 1},
                'RNA': {'C': 5 * units.mg / units.mL},
                'Protein': {'X': 50 * units.mg / units.mL}}}}

    # make a txtl composite, embedded under an 'agents' store
    txtl_composer = AgentDivision({'growth': {'default_growth_noise': 0}})
    txtl_composite1 = txtl_composer.generate(
        {'agent_id': agent_id},
        path=('agents', agent_id))

    # check the agents before division
    for n in ['processes', 'steps', 'topology', 'flow']:
        assert '0' in txtl_composite1[n]['agents'], f'agent 0 not in {n}'

    # make the experiment
    hierarchy_experiment1 = Engine(
        composite=txtl_composite1,
        initial_state=initial_state,
        emit_step=100.0)

    # run the experiment long enough to divide
    hierarchy_experiment1.update(2000)

    # check that the agents updated after division
    for n in ['processes', 'steps', 'topology']:  # TODO: add flow
        assert '0' not in txtl_composite1[n]['agents'], \
            f'agent 0 not removed from {n}'
        assert '00' in txtl_composite1[n]['agents'], \
            f'agent 00 not added to {n}'
        assert '01' in txtl_composite1[n]['agents'], \
            f'agent 01 not added to {n}'


if __name__ == '__main__':
    test_hierarchy_update()



from typing import Dict, Any
from vivarium.core.process import Deriver
from vivarium.library.units import units
from vivarium.processes.tree_mass import AVOGADRO


class CountsToMolar(Deriver):
    """
    Process for deriving molar concentrations from counts
    """

    name = 'counts_to_molar'
    defaults: Dict[str, Any] = {
        'keys': [],
    }

    def initial_state(self, config=None):
        mass = 1000 * units.fg
        density = 1100 * units.g / units.L
        volume = mass / density
        mmol_to_counts = AVOGADRO * volume
        return {
            'global': {
                'volume': volume.to('fL'),
                'mmol_to_counts': mmol_to_counts.to('L/mmol')
            }}

    def ports_schema(self):
        initial_state = self.initial_state()
        return {
            'global': {
                'volume': {
                    '_default': initial_state[
                        'global']['volume'].to('fL')}},
            'counts': {
                concentration: {
                    '_divider': 'split'}
                for concentration in self.parameters['keys']},
            'concentrations': {
                concentration: {
                    '_divider': 'set',
                    '_updater': 'set'}
                for concentration in self.parameters['keys']}}

    def next_update(self, timestep, states):
        volume = states['global']['volume']
        mmol_to_counts = (AVOGADRO * volume).to('L/mmol').magnitude
        counts = states['counts']

        # concentration update
        concentrations = {}
        if volume != 0:
            for molecule, count in counts.items():
                concentrations[molecule] = count / mmol_to_counts
            return {
                'concentrations': concentrations}
        print('volume is 0!')
        return {}


class MolarToCounts(Deriver):
    """
    Process for deriving counts from molar concentration
    """
    name = 'molar_to_counts'
    defaults: Dict[str, Any] = {
        'keys': [],
    }

    def initial_state(self, config=None):
        mass = 1000 * units.fg
        density = 1100 * units.g / units.L
        volume = mass / density
        return {
            'global': {
                'volume': volume.to('fL'),
            }}

    def ports_schema(self):
        initial_state = self.initial_state()
        return {
            'global': {
                'volume': {
                    '_default': initial_state['global']['volume'].to('fL')},
            },
            'counts': {
                molecule: {
                    '_default': 0,
                    '_divider': 'split'}
                for molecule in self.parameters['keys']},
            'concentrations': {
                molecule: {
                    '_default': 0.0}
                for molecule in self.parameters['keys']}}

    def next_update(self, timestep, states):
        volume = states['global']['volume']
        mmol_to_counts = (AVOGADRO * volume).to('L/mmol').magnitude
        concentrations = states['concentrations']  # assumes mmol/L

        counts = {}
        for molecule, concentration in concentrations.items():
            counts[molecule] = {
                '_value': int(concentration * mmol_to_counts),
                '_updater': 'set'}

        return {
            'counts': counts}

def test_derivers():
    pass


if __name__ == '__main__':
    test_derivers()
